/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import com.unitesk.aspectrace.common.CommonAdapter;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Base class for all aspect adapters.
 *
 * @author Sergey Groshev
 */
public abstract class AspectAdapter {

  private List<AspectListener> listeners = new ArrayList<AspectListener>();

  private TraceProvider traceProvider;

  private String currentChannel;

  // ------------------------------------------------------
  // Public interface: configure listeners

  /**
   * Adds listener.
   *
   * @param listener Listener for aspect messages
   * @throws ClassCastException Incorrect listener class
   * @throws IllegalStateException Cannot change the set of listeners now
   */
  public void addListener(AspectListener listener)
      throws ClassCastException, IllegalStateException {
    if (listener == null) {
      throw new NullPointerException();
    }
    if (!getListenerClass().isInstance(listener)) {
      throw new ClassCastException(
          "Invalid listener class: "
              + listener.getClass()
              + " does not implement "
              + getListenerClass());
    }
    listener.attachedToAdapter(getAspect(), this);
    listeners.add(listener);
  }

  /**
   * Deletes handler.
   *
   * @param listener Handler to delete
   * @throws IllegalStateException Cannot change the set of listeners now
   */
  public void removeListener(AspectListener listener) throws IllegalStateException {
    listeners.remove(listener);
  }

  // ------ Methods subclasses MUST implement

  /**
   * Handles one message from provider. Realized by concrete aspect dependent adapters. Usual method
   * behaviour - converts raw message from provider into aspect dependent message and distributes it
   * to handlers. Class messages {@link TagMessage} come in adapter via method call {@link
   * #processTagMessage(TagMessage)}.
   *
   * @throws IllegalArgumentException Incorrect message struct or incorrect message data
   */
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    setCurrentChannel(msg.getChannel());
  }

  /** Gets interface class which must be realized by all handlers of this aspect. */
  public abstract Class<? extends AspectListener> getListenerClass();

  /** Gets the name of the aspect that is being handled. */
  public abstract String getAspect();

  // ------------------------------------------------------
  // Extension interface

  /**
   * Notification about tags change from provider. This message is being send to the CommonAdapter
   * and to the corresponding aspect adapter.
   */
  protected void processTagMessage(TagMessage msg) {
    for (AspectListener lst : (List<AspectListener>) getListeners()) {
      lst.processTagChange(msg);
    }
  }

  /**
   * Gets list of handlers for the internal use. Subclasses must override this method with an
   * appropriate type of return value.
   */
  protected List<? extends AspectListener> getListeners() {
    return listeners;
  }

  /** Gets current provider which current adapter is working with. */
  final void setTraceProvider(TraceProvider traceProvider) {
    this.traceProvider = traceProvider;
  }

  /** Gets current provider. */
  protected final TraceProvider getTraceProvider() {
    return traceProvider;
  }

  public String getCurrentChannel() {
    return currentChannel;
  }

  protected void setCurrentChannel(String currentChannel) {
    this.currentChannel = currentChannel;
  }

  // ------------------------------------------------------
  // Tag management

  /**
   * Adds tag in the aspect. See {@link TraceProvider#addProcessorTag(String, String, String,
   * String)}.
   */
  public void addProcessorTag(String tag, String parent) {
    String channel = null; // TODO: channel
    traceProvider.addProcessorTag(getAspect(), channel, tag, parent);
  }

  /**
   * Deletes tag in the aspect. See {@link TraceProvider#removeProcessorTag(String, String, String,
   * String)}.
   */
  public void removeProcessorTag(String tag, String parent) {
    String channel = null; // TODO: channel
    traceProvider.removeProcessorTag(getAspect(), channel, tag, parent);
  }

  /**
   * Gets copy of all tags from provider. Calls method {@link TraceProvider#getAllTags()}
   *
   * @return Copy of all tags from provider. Changes of return value have no influence on provider.
   */
  public final Map<String, Set<String>> getAllTags() {
    return traceProvider.getAllTags();
  }

  /**
   * Gets tags for the aspect from the provider. Calls method {@link
   * TraceProvider#getTags(String)}
   *
   * @return Direct link on the corresponding part of provider state.
   */
  protected final Set<String> getMyTags() {
    return traceProvider.getTags(getAspect());
  }

  // ------------------------------------------------------
  // Interface for trace providers

  /**
   * Notification from provider: begin trace handling. By agreement, this method is overridden only
   * in the class which {@link CommonAdapter} distributes the corresponding message between the
   * handlers and provider notifies only this method.
   */
  public void processingStarted(String providerDesc) {}

  /**
   * Notification from provider: end trace handling. By agreement, this method is overridden only in
   * the class which {@link CommonAdapter}, distributes the corresponding message between the
   * handlers and provider notifies only this method.
   */
  public void processingFinished(String providerDesc, boolean normally) {}
}
