/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * {@link TraceMessage message} element. The best external representation is an XML-node with
 * corresponding attributes and children.
 */
public final class TraceNode /*implements java.io.Serializable*/ {
  String name; // ==null for text
  String text; // ==null for structured node
  Map<String, String> attrs;
  List<TraceNode> children;

  TraceNode() {}

  public TraceNode(String name) {
    this(name, (Map<String, String>) null, (TraceNode[]) null);
  }

  public TraceNode(String name, String attrName, String attrValue) {
    // Create singleton map for awhile. Will be replaced with HashMap if needed
    this(name, Collections.singletonMap(attrName, attrValue), (TraceNode[]) null);
  }

  public TraceNode(String name, String... attrs) {
    this(name, TraceUtils.createStringMap(attrs));
  }

  /**
   * Main constructor.
   *
   * @param name Node name.
   * @param attrs Attributes ({@code null} is allowed).
   * @param children Child nodes ({@code null} or an empty array are both allowed).
   */
  public TraceNode(String name, Map<String, String> attrs, TraceNode... children) {
    this.name = name;
    this.attrs = attrs;

    if (children != null && children.length > 0) {
      this.children = new ArrayList<TraceNode>(children.length);
      for (int i = 0; i < children.length; i++) {
        this.children.add(children[i]);
      }
    }
  }

  /** Creates a text node. */
  public static TraceNode createTextNode(String text) {
    TraceNode tn = new TraceNode();
    tn.text = text;
    return tn;
  }

  /** Creates with a single text child. */
  public static TraceNode createNodeWithText(String name, String text) {
    return new TraceNode(name, null, createTextNode(text));
  }

  // ------------------------------------------------------

  /**
   * Prepares a node for the further use after setting all attributes and children. Method is called
   * when creating {@link TraceMessage message trace} on the base of this node. Method sets
   * attributes and children to not-{@code null}.
   *
   * <p>It is possible to block node from changes but for the present is turned off because:
   *
   * <ol>
   *   <li>In normal case scenario message is passed only one time (from provider to adapter or from
   *       aspect tracer to main tracer) therefore it is difficult to corrupt data so that it has
   *       consequences. However, this policy may be changed because of the appearance of the
   *       mechanism {@link MessageStream}.
   *   <li>It realizes the mechanism of {@link MessageStreamFilter trace filters} which may need to
   *       modify messages. For these purposes unblocking attributes and child list is allowed using
   *       following methods {@link #setAttr(String, String)}, {@link #removeAttr(String)} и {@link
   *       #addChild(TraceNode)}.
   * </ol>
   */
  protected void setReady() {
    if (attrs == null || attrs.size() == 0) {
      attrs = Collections.emptyMap();
    }

    if (children == null || children.size() == 0) {
      children = Collections.emptyList();
    } else {
      for (TraceNode tn : children) {
        tn.setReady();
      }
    }
  }

  /**
   * Determines whether a node is structured or not.
   *
   * @return {@code true} if the node is structured {@code false} if the node is a text node.
   */
  public boolean isNode() {
    return name != null;
  }

  /**
   * Gets a node name.
   *
   * @return {@code null} for a text node and the name for a structured node.
   */
  public String getName() {
    return name;
  }

  /** Determines whether a node is a text node or not. */
  public boolean isText() {
    return text != null;
  }

  /**
   * Gets a text.
   *
   * @return {@code null} for a structured node and the text for a text node.
   */
  public String getText() {
    return text;
  }

  /** Gets node attributes. */
  public Map<String, String> getAttrs() {
    return attrs;
  }

  /** Gets attribute value. */
  public String getAttr(String attrName) {
    return attrs.get(attrName);
  }

  /** Sets attribute value. */
  public void setAttr(String name, String value) {
    if (attrs == null) {
      attrs = new LinkedHashMap<String, String>();
    }
    try {
      attrs.put(name, value);
    } catch (UnsupportedOperationException e) { // Expected from EmptyMap and SingletonMap
      attrs = new LinkedHashMap<String, String>(attrs);
      attrs.put(name, value);
    }
  }

  /** Deletes attribute. */
  public void removeAttr(String name) {
    if (attrs == null) {
      return;
    }
    try {
      attrs.remove(name);
    } catch (UnsupportedOperationException e) { // Expected from EmptyMap and SingletonMap
      attrs = new LinkedHashMap<String, String>(attrs);
      attrs.remove(name);
    }
  }

  /** Determines whether this node has children or not. */
  public boolean hasChildren() {
    return children != null && children.size() > 0;
  }

  /**
   * Gets a list of children.
   *
   * @return list of children. Always not-{@code null}.
   */
  public List<TraceNode> getChildren() {
    if (children == null) {
      children = Collections.emptyList();
    }
    return children;
  }

  /** Gets a child by his index. */
  public TraceNode getChild(int index) throws IndexOutOfBoundsException {
    return children.get(index);
  }

  /**
   * Adds a node to the end of child list. It is recommended to use this method instead of {@code
   * getChildren().add(...)}
   */
  public void addChild(TraceNode child) {
    if (children == null) {
      children = new ArrayList<TraceNode>();
    }
    try {
      children.add(child);
    } catch (UnsupportedOperationException e) { // Expected from EmptyList
      ArrayList<TraceNode> ch = new ArrayList<TraceNode>(children.size() + 4);
      if (!children.isEmpty()) {
        ch.addAll(children);
      }
      ch.add(child);
      children = ch;
    }
  }

  /** Adds a text child. */
  public void addTextChild(String text) {
    addChild(createTextNode(text));
  }

  /* For debug purposes.*/
  @Override
  public String toString() {
    return "TraceNode " + name + ' ' + attrs + ' ' + text;
  }
}
