/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import com.unitesk.aspectrace.common.CommonListener;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Trace provider. Represents a stream of {@link AspectMessage aspect messages}, distributed between
 * corresponding adapters in the following way:
 *
 * <ul>
 *   <li>Begin/end trace messages are directed to an empty aspect ("").
 *   <li>Tag messages are directed to an empty aspect and to a parent aspect via corresponding
 *       adapters аспекту и породившему их аспекту, after that are distributed between listeners via
 *       methods {@link CommonListener#processGlobalTagChange(TagMessage)} and {@link
 *       AspectListener#processTagChange(TagMessage)}.
 *   <li>All aspect-dependent messages (including messages from a base aspect) are directed to the
 *       corresponding adapter.<br>
 *       <i>Note: Error messages and coverage messages independently of a logically connected aspect
 *       belong to "error" aspect and "coverage" aspect respectively.</i>
 * </ul>
 *
 * <p>If a trace provider realization suggests a work in it's own active control stream in which
 * messages are pulled from external sources (model <b>PULL</b>) then method {@link #process()} is
 * redefined in the corresponding class. In this scenario the base struct of the class provider is
 * the following:
 *
 * <pre>
 * public void process() { // Called from TraceProcessorManager.process()
 *   fireTraceStart(...);
 *   while(messages) {
 *     fireMessage(...); or fireTagMessage(...);
 *   }
 *   fireTraceEnd(...);
 * }
 * </pre>
 *
 * <p>If a trace provider realization suggests a work in the other's active control stream in which
 * messages are pushed by external sources (model <b>PUSH</b>) then the base struct of the class is
 * the following:
 *
 * <pre>
 * TraceProcessorManager tpm; // Gets from the outside
 * stream_start_event_handler() {
 *   attachToManager(tpm);
 *   fireTraceStart(...);
 * }
 * next_external_event_handler() {
 *   fireMessage(...); or fireTagMessage(...);
 * }
 * end_stream_event_handler() {
 *   fireTraceEnd(...);
 * }
 * </pre>
 *
 * <p>Note: The class {@code TraceProvider} has the realizations of methods {@link
 * #fireTraceStart(String)}, {@link #fireMessage(TraceMessage)}, {@link #fireTagMessage(TagMessage)}
 * and {@link #fireTraceEnd(String, boolean)} which are {@code synchronized} but not {@code final}.
 *
 * @author Sergey Groshev
 */
public abstract class TraceProvider {
  // TODO: Maybe {@link MessageStreamFilter фильтры} will be inserted.

  // TODO? Move all work with adapters, listeners, exception handling policy and tags into the
  // TraceProcessorManagerClass.

  /**
   * Map: aspect +> set of active tags. Tag set is ordered <b>inside each aspect</b>: iteration
   * order must coincide with the order of tag addition.
   */
  private Map<String, LinkedHashSet<String>> activeTags =
      new HashMap<String, LinkedHashSet<String>>();

  private Map<String, AspectAdapter> adapters = new HashMap<String, AspectAdapter>();

  private boolean isInProcessing = false;
  private List<TagMessage> pendingTags = new ArrayList<TagMessage>();

  protected TraceProvider() {}

  /**
   * Begins message handling loop if this is an active trace source (model PULL). If this trace
   * source doesn't get messages in his own active stream but works as a message converter from the
   * others' active (PUSH) sources then - empty realization by default.
   *
   * @throws IOException Input-Output error
   * @throws IllegalStateException The handling has already begun
   * @throws IllegalArgumentException Trace error (casted by a trace source, by an aspect adapter or
   *     by a trace handler)
   */
  public void process() throws IOException, IllegalStateException, IllegalArgumentException {}

  /**
   * Connects all adapters from a specified trace handler manager to itself. Calls {@link
   * TraceProcessorManager#attachAdapters(TraceProvider) tpm.attachAdapters(this)}. This method
   * calls providers which DON'T work in an active stream of the corresponding trace handle manager
   * in the beginning of a trace.
   *
   * @see #fireTraceStart(String)
   */
  protected void attachToManager(TraceProcessorManager tpm) {
    tpm.attachAdapters(this);
  }

  // ------------------------------------------------------
  // Adapters

  /**
   * Sets an adapter for an aspect. The adapter can be used multiple times.
   *
   * @throws IllegalArgumentException Incorrect adapter or this aspect has got already a different
   *     adapter instance.
   * @throws IllegalStateException The handling has already begun
   */
  protected final void setAdapter(String aspect, AspectAdapter adapter)
      throws IllegalArgumentException, IllegalStateException {
    if (!aspect.equals(adapter.getAspect())) {
      throw new IllegalArgumentException("Attempt to set adapter for wrong aspect");
    }
    AspectAdapter oldAdapter = adapters.get(aspect);
    if (oldAdapter == adapter) {
      return;
    }
    if (oldAdapter != null) {
      throw new IllegalArgumentException("Another adapter for aspect " + aspect + " already set");
    }
    adapters.put(aspect, adapter);
    adapter.setTraceProvider(this);
  }

  /**
   * Gets an adapter for an aspect. Unlike {@link TraceProcessorManager#getAdapter(String)} this
   * method doesn't create adapters.
   *
   * @return Adapter or null
   */
  public final AspectAdapter getAdapter(String aspect) {
    return adapters.get(aspect);
  }

  /** Removes aspect adapter. */
  public final void removeAdapter(String aspect) {
    adapters.remove(aspect);
  }

  // ------------------------------------------------------
  // Messages

  /** Notifies empty aspect handlers about the trace handle start. */
  protected synchronized void fireTraceStart(String traceDesc) {
    AspectAdapter defAd = getAdapter("");
    if (defAd != null) {
      defAd.processingStarted(traceDesc);
    }
  }

  /** Notifies empty aspect handlers about the trace handle end. */
  protected synchronized void fireTraceEnd(String traceDesc, boolean verdict) {
    AspectAdapter defAd = getAdapter("");
    if (defAd != null) {
      defAd.processingFinished(traceDesc, verdict);
    }
  }

  protected final void fireMessage(String aspect, String channel, TraceNode tn)
      throws IllegalArgumentException {
    fireMessage(new TraceMessage(aspect, channel, tn));
  }

  /**
   * Distributes a trace message via the corresponding {@link AspectAdapter adapter}.
   * {@link TagMessage change tag} messages must be distributed via {@link
   * #fireTagMessage(TagMessage)} method.
   *
   * @throws IllegalArgumentException Aspect adapter has found an error in message.
   */
  protected synchronized void fireMessage(TraceMessage msg)
      throws IllegalArgumentException {
    isInProcessing = true;
    try {
      AspectAdapter adapter = adapters.get(msg.getAspect());
      if (adapter == null) {
        return;
      }

      adapter.processMessage(msg);

      while (!pendingTags.isEmpty()) {
        // Tag dispatching may produce additional deferred tag events.
        // Hope, they will not cause infinite trigger loop.
        TagMessage[] msgs = pendingTags.toArray(new TagMessage[pendingTags.size()]);
        pendingTags.clear();
        for (int i = 0; i < msgs.length; i++) {
          doFireTag(msgs[i]);
        }
      }
    } finally {
      isInProcessing = false;
    }
  }

  /**
   * Distributes a tag change message or a synchronized message via adapters.
   * Message is distributed via an empty aspect adapter and via an adapter corresponding aspect tag.
   * If this method is called during other's message handling then the corresponding
   * tag message is delayed until current handling end.
   */
  protected synchronized void fireTagMessage(TagMessage msg)
      throws IllegalArgumentException {
    if (isInProcessing) {
      pendingTags.add(msg);
      return;
    }
    doFireTag(msg);
  }

  protected final void fireSyncMessage(String aspect, String channel, String name)
      throws IllegalArgumentException {
    fireTagMessage(new TagMessage(aspect, channel, name, null, false, true, false));
  }

  private void doFireTag(TagMessage msg) throws IllegalArgumentException {
    String aspect = msg.getAspect();
    String tag = msg.getTag();

    if (!msg.isSync()) {
      // TODO: parent, channel
      if (msg.isStart()) {
        getTags(aspect).add(tag);
      } else {
        getTags(aspect).remove(tag);
      }
    }

    doFireTag0("", msg);
    if (aspect != null && aspect.length() != 0) {
      doFireTag0(aspect, msg);
    }
  }

  private void doFireTag0(String aspect, TagMessage msg) {
    AspectAdapter adapter = adapters.get(aspect);
    if (adapter != null) {
      adapter.processTagMessage(msg);
    }
  }

  // ------------------------------------------------------
  // Tags

  /**
   * Adds a tag from a trace. Message is distributed in the standard way via method {@link
   * #fireTagMessage(TagMessage)}
   */
  protected final void addProviderTag(String aspect, String channel, String tag, String parent) {
    TagMessage msg = new TagMessage(aspect, channel, tag, parent, true, false, false);
    fireTagMessage(msg);
  }

  /**
   * Deletes a tag from a trace. Message is distributed in the standard way via method {@link
   * #fireTagMessage(TagMessage)}
   */
  protected final void removeProviderTag(String aspect, String channel, String tag, String parent) {
    TagMessage msg = new TagMessage(aspect, channel, tag, parent, false, false, false);
    fireTagMessage(msg);
  }

  /**
   * Adds a user tag. The method is usually called by a trace handler.
   * A change tag message with the flag
   * isUser=true is distributed in the standard way via method {@link
   * #fireTagMessage(TagMessage)}
   */
  protected final void addProcessorTag(String aspect, String channel, String tag, String parent) {
    TagMessage msg = new TagMessage(aspect, channel, tag, parent, true, false, true);
    fireTagMessage(msg);
  }

  /**
   * Deletes a user tag. The method is usually called by a trace handler.
   * A change tag message with the flag
   * isUser=true is distributed in the standard way via method {@link
   * #fireTagMessage(TagMessage)}
   */
  protected final void removeProcessorTag(
      String aspect, String channel, String tag, String parent) {
    TagMessage msg = new TagMessage(aspect, channel, tag, parent, false, false, true);
    fireTagMessage(msg);
  }

  /**
   * Gets tags for an aspect.
   *
   * @return A set of tags. A return value directly refers to the current provider state.
   */
  protected Set<String> getTags(String aspect) {
    // TODO: parent, channel

    LinkedHashSet<String> tags = activeTags.get(aspect);
    if (tags == null) {
      tags = new LinkedHashSet<String>();
      activeTags.put(aspect, tags);
    }
    return tags;
  }

  /**
   * Gets all tags.
   *
   * @return A copy of the complete tag set for all aspects. Mapping values are
   *     class {@link LinkedHashSet} instances which guarantees
   *     the order of iteration corresponding with the order of tag creation (in trace).
   *     In case of a inner tag outer tag appears first.
   */
  protected Map<String, Set<String>> getAllTags() {
    Map<String, Set<String>> allTags = new HashMap<String, Set<String>>(activeTags.size(), 1.0f);
    for (Map.Entry<String, LinkedHashSet<String>> me : activeTags.entrySet()) {
      LinkedHashSet<String> val = me.getValue();
      if (val.isEmpty()) {
        continue;
      }
      allTags.put(me.getKey(), (Set<String>) val.clone());
    }
    return allTags;
  }
}
