/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace;

import com.unitesk.aspectrace.tracer.TraceFileWriter;
import com.unitesk.aspectrace.tracer.TraceToProcessor;
import com.unitesk.aspectrace.tracer.XmlFormatter;

import java.io.File;
import java.io.IOException;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.WeakHashMap;

/**
 * Central control point of Aspectrace tracing.
 *
 * @author Sergey Groshev
 */
public final class Tracer implements Cloneable {

  public static final String PROP_DEF_CHANNEL = "default.channel";
  public static final String PROP_FILE_ENCODING = "file.encoding";

  public static final String DEFAULT_CHANNEL = "0";
  public static final String DEFAULT_ENCODING = "windows-1251";

  private static final Properties defProps = new Properties();

  static {
    defProps.put(PROP_DEF_CHANNEL, DEFAULT_CHANNEL);
    defProps.put(PROP_FILE_ENCODING, DEFAULT_ENCODING);
  }

  // ------------------------------------------------------
  // Controls tracer instances

  private String channelId;

  private Tracer() {
    channelId = allocateChannelId();
  }

  private static Tracer masterTracer;

  /**
   * A weak reference for a thread. After thread's death it gets to the queue for a tracer cleaning.
   * Until then supports a live link to the tracer.
   */
  static class ThreadRef extends WeakReference<Thread> {
    Tracer tr;

    ThreadRef(Thread thr, Tracer tr) {
      super(thr, queue);
      this.tr = tr;
    }
  }

  /**
   * List of existing (not necessarily active) threads.
   * In process of the threads dying is cleaned up
   * via WeakHashMap.
   */
  private static WeakHashMap<Thread, ThreadRef> tracers = new WeakHashMap<Thread, ThreadRef>();

  private static ReferenceQueue<Thread> queue = new ReferenceQueue<Thread>();

  /**
   * Gets a tracer instance for the current process/stream/etc.
   *
   * @see #endTrace()
   */
  public static Tracer getInstance() {
    // Looks through the queue of the dead threads and cleans up the tracer.
    // Supposing WeakHashMap keys will be cleaned up by themselves.
    Reference<? extends Thread> ref;
    while ((ref = queue.poll()) != null) {
      tracers.size(); // Calls expungeStaleEntries()
      Tracer tr = ((ThreadRef) ref).tr;
      if (tr != masterTracer) {
        tr.endTrace();
      }
    }

    Thread currentThread = Thread.currentThread();
    ThreadRef tref = tracers.get(currentThread);
    if (tref == null) {
      tref = new ThreadRef(currentThread, new Tracer());
      synchronized (tracers) {
        tracers.put(currentThread, tref);
      }
    }
    return tref.tr;
  }

  /**
   * Sets the tracer of the current stream as master.
   * In this case tracer instances from other streams
   * which tracer channels haven't been explicitly set will use the channels from this tracer.
   *
   * <p>If there has been made a call in a thread manually then
   * {@link #endTrace()} call must be made manually afterwards as well.
   */
  public static void setMasterTracer() {
    masterTracer = getInstance();
  }

  // Second to last chance to end the tracing.
  @Override
  protected void finalize() {
    endTrace();
  }

  static {
    // Last chance to end the tracing.
    Runtime.getRuntime()
        .addShutdownHook(
            new Thread(
                new Runnable() {
                  @Override
                  public void run() {
                    synchronized (tracers) {
                      for (ThreadRef tref : tracers.values()) {
                        tref.tr.endTrace();
                      }
                    }
                  }
                }));
  }

  // ------------------------------------------------------
  // Channels

  /** Gets a channel ID for the current process/stream/etc. */
  public static String getChannel() {
    return getInstance().get_channel();
  }

  /** Gets a channel ID which is processed by the current tracer. */
  public String get_channel() {
    return channelId;
  }

  public static void setChannel(String ch) {
    getInstance().channelId = ch;
  }

  static int channelIdCounter = 0;

  public static synchronized String allocateChannelId() {
    return String.valueOf(channelIdCounter++);
  }

  // ------------------------------------------------------
  // Configurations of formats and assign trace points

  /** Trace stream. Maybe more than one. */
  public interface Channel {
    /**
     * Starts trace.
     *
     * @param props Conf.
     * @throws IOException Input-output error.
     * @throws IllegalStateException Impossible to start trace, trace has already begun or
     *     conf error.
     * @throws IllegalArgumentException Invalid properties in {@code props}.
     */
    public void startTrace(Properties props)
        throws IOException, IllegalStateException, IllegalArgumentException;

    /**
     * Processes trace message.
     *
     * @param msg Trace message. Realizations must recognize following message types: {@link
     *     TraceMessage}, {@link TagMessage}, {@link DebugMessage}
     * @throws IOException Input-output error.
     * @throws IllegalStateException Trace has not begun yet or conf error.
     * @throws IllegalArgumentException Configuration internal problems.
     */
    public void dispatch(AbstractTraceMessage msg)
        throws IOException, IllegalStateException, IllegalArgumentException;

    /**
     * Ends a trace.
     *
     * @throws IOException Input-output error.
     * @throws IllegalStateException Trace has not begun yet or conf error.
     * @throws IllegalArgumentException Configuration internal problems.
     */
    public void endTrace() throws IOException, IllegalStateException, IllegalArgumentException;
  }

  /** Formats trace messages into the external string representation. */
  public interface Formatter {
    /**
     * Initializes a formatting.
     *
     * @param props Conf properties.
     * @return Prologue string for a {@link Writer trace writer}.
     */
    public String formatTraceStart(Properties props);

    /** Formats a regular message. */
    public String formatMessage(TraceMessage msg);

    /** Formats a tag message. */
    public String formatTagMessage(TagMessage msg);

    /**
     * Formats a debug message (comment).
     * It is possible to format into an empty string so it will be lost afterwards.
     */
    public String formatDebugMessage(DebugMessage msg);

    /** Formats an epilogue for {@link Writer trace writer}. */
    public String formatTraceEnd();
  }

  /** Writes messages formatted via {@link Tracer.Formatter}. */
  public interface Writer {
    /**
     * Begins trace writing.
     *
     * @param prolog Prologue string.
     * @param props Conf properties.
     */
    public void startTrace(String prolog, Properties props)
        throws IOException, IllegalStateException, IllegalArgumentException;

    public void writeMessage(String msg)
        throws IOException, IllegalStateException, IllegalArgumentException;

    public void endTrace(String epilog)
        throws IOException, IllegalStateException, IllegalArgumentException;
  }

  /**
   * Trace standard configuration: formats messages via {@link Tracer.Formatter}
   * into the stream of symbols and writes them down in {@link Tracer.Writer}.
   */
  public static final class FormattedChannel implements Channel {
    Formatter formatter;
    Writer writer;

    public FormattedChannel(Formatter formatter, Writer writer) {
      this.formatter = formatter;
      this.writer = writer;
    }

    @Override
    public void startTrace(Properties props)
        throws IOException, IllegalStateException, IllegalArgumentException {
      writer.startTrace(formatter.formatTraceStart(props), props);
    }

    @Override
    public void dispatch(AbstractTraceMessage msg)
        throws IOException, IllegalStateException, IllegalArgumentException {
      writer.writeMessage(format(formatter, msg));
    }

    @Override
    public void endTrace() throws IOException, IllegalStateException, IllegalArgumentException {
      writer.endTrace(formatter.formatTraceEnd());
    }
  }

  static String format(Formatter frm, AbstractTraceMessage msg) {
    switch (msg.getKind()) {
      case TRACE_MESSAGE:
        return frm.formatMessage((TraceMessage) msg);
      case TAG_MESSAGE:
        return frm.formatTagMessage((TagMessage) msg);
      case DEBUG_MESSAGE:
        return frm.formatDebugMessage((DebugMessage) msg);
      default: // Should not be
        throw new ClassCastException();
    }
  }

  // ------------------------------------------------------

  private List<Channel> channels = new ArrayList<Channel>();
  private List<MessageStreamFilter> filters = new ArrayList<MessageStreamFilter>();

  /** Adds a trace standard configuration.*/
  public void addTracing(Formatter frm, Writer wrt, Properties props)
      throws IllegalStateException, IllegalArgumentException, IOException {
    addTracing(new FormattedChannel(frm, wrt), props);
  }

  /** Adds a trace standard configuration with the default properties. */
  public void addTracing(Formatter frm, Writer wrt)
      throws IllegalStateException, IllegalArgumentException, IOException {
    // defProps is wrapped to avoid modification by user-supplied f and w
    addTracing(new FormattedChannel(frm, wrt), new Properties(defProps));
  }

  /** Adds a trace random configuration with the default properties. */
  public void addTracing(Channel chn)
      throws IllegalStateException, IllegalArgumentException, IOException {
    // defProps is wrapped to avoid modification by user-supplied Channel
    addTracing(chn, new Properties(defProps));
  }

  /** Adds a trace random configuration and starts it. */
  public synchronized void addTracing(Channel chn, Properties props)
      throws IllegalStateException, IllegalArgumentException, IOException {
    chn.startTrace(props);
    channels.add(chn);
  }

  /** Adds a trace in a XML-file with the default coding. */
  public void addXmlTrace(File file)
      throws IOException, IllegalStateException, IllegalArgumentException {
    addTracing(new XmlFormatter(), new TraceFileWriter(file), defProps);
  }

  /** Adds a trace in a XML-file with a given coding. */
  public void addXmlTrace(File file, String encoding)
      throws IOException, IllegalStateException, IllegalArgumentException {
    Properties props = new Properties(defProps);
    props.setProperty(PROP_FILE_ENCODING, encoding);
    addTracing(new XmlFormatter(), new TraceFileWriter(file), props);
  }

  /**
   * Adds a trace in a template XML-file with the default coding.
   */
  public void addXmlTrace(String pattern)
      throws IOException, IllegalStateException, IllegalArgumentException {
    addXmlTrace(TraceUtils.resolveFilePattern(pattern));
  }

  /** Adds a trace in a template XML-file with a given coding. */
  public void addXmlTrace(String pattern, String encoding)
      throws IOException, IllegalStateException, IllegalArgumentException {
    addXmlTrace(TraceUtils.resolveFilePattern(pattern), encoding);
  }

  /**
   * Adds a trace in a TraceToProcessor.
   *
   * @param manager Defined configuration of handlers. All necessary handlers and adapters
   *               must be configured and connected before calling this method.
   * @throws IllegalArgumentException Manager error.
   * @throws IllegalStateException Invalid state of manager or tracer.
   */
  public void addProcessor(TraceProcessorManager manager)
      throws IllegalArgumentException, IllegalStateException, IOException {
    addTracing(new TraceToProcessor(manager));
  }

  /**
   * Adds a filter through which all of the tracing will go.
   * Filters may be more than one - in that case messages will
   * go through all of them in the order of filter adding.
   */
  public void addMessageStreamFilter(MessageStreamFilter messageStreamFilter) {
    filters.add(messageStreamFilter);
  }

  /**
   * Removes a trace filter.
   *
   * @see #addMessageStreamFilter(MessageStreamFilter)
   */
  public void removeMessageStreamFilter(MessageStreamFilter messageStreamFilter) {
    filters.remove(messageStreamFilter);
  }

  // ------------------------------------------------------
  // Tracing

  private Map<Formatter, String> formatCache = new HashMap<Formatter, String>();

  /** Traces a message. */
  public synchronized void trace_message(AbstractTraceMessage msg) {
    if (channels.size() == 0) {
      if (masterTracer != null && masterTracer != this) {
        masterTracer.trace_message(msg);
      }
      return;
    }

    msg = TraceUtils.applyFilters(filters, msg);
    if (msg == null) {
      return;
    }

    for (Channel chn : channels) {
      try {
        if (chn instanceof FormattedChannel) { // Usual case
          FormattedChannel fc = (FormattedChannel) chn;
          String str = formatCache.get(fc.formatter);
          if (str == null) {
            str = format(fc.formatter, msg);
            formatCache.put(fc.formatter, str);
          }
          fc.writer.writeMessage(str);
        } else { // Special cases
          chn.dispatch(msg);
        }
      } catch (Throwable e) {
        tracingError(e);
      }
    }

    formatCache.clear();
  }

  /** Traces an open tag message. */
  public void trace_tag_start(String aspect, String tag, String parent) {
    trace_message(new TagMessage(aspect, getChannel(), tag, parent, true, false, false));
  }

  /** Traces a close tag message. */
  public void trace_tag_end(String aspect, String tag, String parent) {
    trace_message(new TagMessage(aspect, getChannel(), tag, parent, false, false, false));
  }

  /** Traces a synchronized tag message. */
  public void trace_sync_tag(String aspect, String name) {
    trace_message(new TagMessage(aspect, getChannel(), name, null, false, true, false));
  }

  /** Traces a debug message. */
  public void trace_debug(String text) {
    trace_message(new DebugMessage(text));
  }

  /**
   * Ends a trace. All errors in channels are ignored.
   *
   * <p>This method is recommended to call in the threads which are doing tracing before their end.
   * If it hasn't been done then the tracing system will try to do so automatically via
   * JVM garbage collector and before the program end but this is not so safe.
   */
  public synchronized void endTrace() {
    for (Channel chn : channels) {
      try {
        chn.endTrace();
      } catch (Throwable e) {
        tracingError(e);
      }
    }
    channels.clear();
  }

  // ------------------------------------------------------
  // Static interface

  /** Traces a message via the default instance. */
  public static void traceMessage(AbstractTraceMessage msg) {
    getInstance().trace_message(msg);
  }

  /** Traces a debug message via the default instance. */
  public static void traceDebug(String text) {
    getInstance().trace_debug(text);
  }

  /** Traces an open tag message via the default instance. */
  public static void traceTagStart(String aspect, String tag, String parent) {
    getInstance().trace_tag_start(aspect, tag, parent);
  }

  /** Traces a close tag message via the default instance. */
  public static void traceTagEnd(String aspect, String tag, String parent) {
    getInstance().trace_tag_end(aspect, tag, parent);
  }

  /** Traces a synchronized tag message via the default instance. */
  public static void traceSyncTag(String aspect, String name) {
    getInstance().trace_sync_tag(aspect, name);
  }

  /** Traces an error in trace procedures. */
  public static void tracingError(Throwable exception) {
    exception.printStackTrace();
  }
}
