/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.common;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.TraceMessage;

import java.util.List;

/**
 * An adapter for an empty aspect.
 *
 * <p>This adapter handles special messages (without namespace in XML).
 * Also it handles tag messages in all aspects by distributing them via {@link
 * CommonListener#processGlobalTagChange(TagMessage) processGlobalTagChange} method of handlers.
 *
 * <p>Method {@link #getAspect} returns an empty string.
 *
 * @author Sergey Groshev
 */
public class CommonAdapter extends AspectAdapter {

  @Override
  public String getAspect() {
    return "";
  }

  @Override
  public Class<CommonListener> getListenerClass() {
    return CommonListener.class;
  }

  @Override
  public void processingStarted(String providerDesc) {
    for (CommonListener lst : getListeners()) {
      lst.processingStarted(providerDesc);
    }
  }

  @Override
  public void processingFinished(String providerDesc, boolean normally) {
    for (CommonListener lst : getListeners()) {
      lst.processingFinished(providerDesc, normally);
    }
  }

  @Override
  protected void processTagMessage(TagMessage msg) {
    if (msg.getAspect().equals("")) { // Tag for default aspect. Normally, this should not happen.
      super.processTagMessage(msg);
    } else { // Notify tag changes for other aspects.
      for (CommonListener lst : getListeners()) {
        lst.processGlobalTagChange(msg);
      }
    }
  }

  @Override
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    super.processMessage(msg);

    String msgName = msg.getName();
    if (msgName.equals("channel")) {
      String id = msg.getAttr("id");
      String desc = msg.getAttr("desc");
      ChannelMessage chMsg = new ChannelMessage(this, id, desc);
      for (CommonListener lst : getListeners()) {
        lst.processChannelMessage(chMsg);
      }
    } else {
      throw new IllegalArgumentException("Unsupported default aspect node: " + msgName);
    }
  }

  @Override
  protected List<CommonListener> getListeners() {
    return (List<CommonListener>) super.getListeners();
  }
}
