/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.coverage;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.LazyNode;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageStructure;

import java.util.ArrayList;
import java.util.List;

/**
 * Adapter for the aspect "coverage".
 *
 * @author Sergey Groshev
 */
public class CoverageAdapter extends AspectAdapter {

  @Override
  public String getAspect() {
    return "coverage";
  }

  @Override
  public Class<CoverageListener> getListenerClass() {
    return CoverageListener.class;
  }

  @Override
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    super.processMessage(msg);

    // LazyNode usage example.
    LazyNode lazyNode = msg.getLazyNode();
    if (lazyNode != null) {
      CoverageStructure str = ((CoverageTracer.CoverageStructureWrapper) lazyNode).cs;
      dispatchStructure(str);
      return;
    }

    // TraceNode normal use.
    String msgName = msg.getName();
    if (msgName.equals("structure")) {
      CoverageStructure str =
          new CoverageStructure(
              msg.getAttr("aspect"), msg.getAttr("name"), buildSubElements(msg.getNode()));
      dispatchStructure(str);
    } else if (msgName.equals("element")) {
      CoverageMessage covMsg = new CoverageMessage(this);
      covMsg.targetAspect = msg.getAttr("aspect");
      covMsg.structName = msg.getAttr("cs");
      List<String> elems = new ArrayList<String>();
      for (TraceNode n = msg.getNode(); n != null; ) {
        if (!"element".equals(n.getName())) {
          throw new IllegalArgumentException(
            "Unsupported coverage element subnode: " + n.getName());
        }
        elems.add(n.getAttr("name"));
        switch (n.getChildren().size()) {
          case 0:
            n = null;
            break;
          case 1:
            n = n.getChild(0);
            break;
          default:
            throw new IllegalArgumentException("Incorrect coverage element structure");
        }
      }
      covMsg.elementChain = elems.toArray(new String[elems.size()]);
      for (CoverageListener lst : getListeners()) {
        lst.processCoverageMessage(covMsg);
      }
    } else if (msgName.equals("impossible")) {
      // TODO !!!
    } else {
      throw new IllegalArgumentException("Unsupported coverage node: " + msgName);
    }
  }

  void dispatchStructure(CoverageStructure str) {
    CoverageStructureMessage msg = new CoverageStructureMessage(this);
    msg.str = str;
    for (CoverageListener lst : getListeners()) {
      lst.processCoverageStructure(msg);
    }
  }

  private static List<CoverageElement> buildSubElements(TraceNode node) {
    List<CoverageElement> list = new ArrayList<CoverageElement>(node.getChildren().size());
    for (TraceNode child : node.getChildren()) {
      if (!"element".equals(child.getName())) {
        throw new IllegalArgumentException(
          "Unsupported coverage structure subnode: " + child.getName());
      }
      CoverageElement ce =
          new CoverageElement(
              child.getAttr("name"),
              Boolean.parseBoolean(child.getAttr("isLeaf")),
              buildSubElements(child));
      list.add(ce);
    }
    return list;
  }

  @Override
  protected List<CoverageListener> getListeners() {
    return (List<CoverageListener>) super.getListeners();
  }
}
