/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.error;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;

import java.util.List;

/**
 * Adapter for "error" aspect.
 *
 * @author Sergey Groshev
 */
public class ErrorAdapter extends AspectAdapter {

  @Override
  public String getAspect() {
    return "error";
  }

  @Override
  public Class<ErrorListener> getListenerClass() {
    return ErrorListener.class;
  }

  @Override
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    super.processMessage(msg);

    String msgName = msg.getName();
    if (msgName.equals("failure")) {
      FailureMessage flrMsg = new FailureMessage(this);
      flrMsg.originAspect = msg.getAttr("aspect");
      for (TraceNode n : msg.getChildren()) {
        if (!n.isNode()) {
          throw new IllegalArgumentException("Plain text child of failure node");
        }
        String childName = n.getName();
        if (childName.equals("short-desc")) {
          if (n.hasChildren()) {
            flrMsg.shortDesc = n.getChild(0).getText();
          } else {
            flrMsg.shortDesc = "";
          }
        } else if (childName.equals("description")) {
          if (n.hasChildren()) {
            flrMsg.fullDesc = n.getChild(0).getText();
          } else {
            flrMsg.fullDesc = "";
          }
        } else if (childName.equals("stack")) {
          int size = n.getChildren().size();
          flrMsg.stack = new StackTraceElement[size];
          for (int i = 0; i < size; i++) {
            TraceNode fn = n.getChildren().get(i);
            if (!"frame".equals(fn.getName())) {
              throw new IllegalArgumentException("frame node expected");
            }
            int line;
            try {
              line = Integer.parseInt(fn.getAttr("line"));
            } catch (NumberFormatException e) {
              line = -1;
            }
            flrMsg.stack[i] =
                new StackTraceElement(
                    fn.getAttr("class"), fn.getAttr("method"), fn.getAttr("source"), line);
          }
        }
        // else: silently ignore unrecognized nodes
      }
      if (flrMsg.stack == null) {
        flrMsg.stack = new StackTraceElement[0];
      }

      for (ErrorListener lst : getListeners()) {
        lst.processFailure(flrMsg);
      }
    } else {
      throw new IllegalArgumentException("Unsupported error node: " + msgName);
    }
  }

  @Override
  protected List<ErrorListener> getListeners() {
    return (List<ErrorListener>) super.getListeners();
  }
}
