/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.error;

import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.Tracer;

/**
 * Tracer for "error" aspect.
 *
 * @author Sergey Groshev
 */
public class ErrorTracer {

  public static final String ASPECT = "error";

  /**
   * Traces error message.
   *
   * @param data Error additional data. If {@link Throwable exception} is passed then its stack is traced
   *     otherwise stack which is one level above.
   * @see #traceFailure(String, String, String, Object, int)
   */
  public static void traceFailure(String aspect, String shortDesc, String fullDesc, Object data) {
    traceFailure(aspect, shortDesc, fullDesc, data, data instanceof Throwable ? 0 : 2);
  }

  /**
   * Traces error message.
   *
   * @param aspect Aspect informing about error.
   * @param shortDesc Short description.
   * @param fullDesc Full description.
   * @param data Error additional data. May be {@link Throwable exception} - in that case
   *     its stack is traced.
   * @param skipFrames Number of the last call stack frames which will be skipped while tracing.
   *     Use negative values if you don't want to trace stack; 0 if you want to trace everything
   *     including this call; 1 - if you want to skip frame corresponding to this method call.
   *     Bigger values - if you want to skip more transitional methods.
   */
  public static void traceFailure(
      String aspect, String shortDesc, String fullDesc, Object data, int skipFrames) {
    // TODO: data
    TraceNode tn = new TraceNode("failure", "aspect", aspect);
    if (shortDesc != null) {
      tn.addChild(TraceNode.createNodeWithText("short-desc", shortDesc));
    }
    if (fullDesc == null && data != null) {
      fullDesc = data.toString();
    }
    if (fullDesc != null) {
      tn.addChild(TraceNode.createNodeWithText("description", fullDesc));
    }
    if (skipFrames >= 0) {
      Throwable th = data instanceof Throwable ? (Throwable) data : new Throwable();
      StackTraceElement[] stackTrace = th.getStackTrace();
      TraceNode stackNode = new TraceNode("stack");
      for (int i = skipFrames; i < stackTrace.length; i++) {
        StackTraceElement elem = stackTrace[i];
        TraceNode trN =
            new TraceNode("frame", "class", elem.getClassName(), "method", elem.getMethodName());
        if (elem.getFileName() != null) {
          trN.setAttr("source", elem.getFileName());
        }
        if (elem.getLineNumber() >= 0) {
          trN.setAttr("line", Integer.toString(elem.getLineNumber()));
        }
        stackNode.addChild(trN);
      }
      tn.addChild(stackNode);
    }

    TraceMessage msg = new TraceMessage(ASPECT, tn);
    Tracer.traceMessage(msg);
  }
}
