/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.fsm;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.MessageLevel;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Adapter for aspect "fsm".
 * Used only for information about UniTESK scenario state machine manager for now.
 *
 * @author Sergey Groshev
 */
public class FsmAdapter extends AspectAdapter {

  @Override
  public String getAspect() {
    return "fsm";
  }

  @Override
  public Class<FsmListener> getListenerClass() {
    return FsmListener.class;
  }

  @Override
  protected void processMessage(TraceMessage msg) throws IllegalArgumentException {
    super.processMessage(msg);

    String msgName = msg.getName();
    if (msgName.equals("start")) {
      FsmStartMessage fsMsg = new FsmStartMessage(this);
      fsMsg.name = msg.getAttr("name");
      fsMsg.fullName = msg.getAttr("full-name");
      for (FsmListener lst : getListeners()) {
        lst.processFsmStart(fsMsg);
      }
    } else if (msgName.equals("end")) {
      FsmEndMessage fsMsg = new FsmEndMessage(this);
      fsMsg.name = msg.getAttr("name");
      for (FsmListener lst : getListeners()) {
        lst.processFsmEnd(fsMsg);
      }
    } else if (msgName.equals("state")) {
      FsmStateMessage fsMsg = new FsmStateMessage(this);
      fsMsg.fsm = msg.getAttr("fsm");
      fsMsg.id = msg.getAttr("id");
      fsMsg.name = msg.getAttr("name");
      for (FsmListener lst : getListeners()) {
        lst.processStateMessage(fsMsg);
      }
    } else if (msgName.equals("start-transition")) {
      TransitionMessage trMsg = new TransitionMessage(this);
      trMsg.isStart = true;
      trMsg.name = msg.getAttr("name");
      trMsg.fsm = msg.getAttr("fsm");
      trMsg.id = msg.getAttr("id");
      for (FsmListener lst : getListeners()) {
        lst.processTransitionMessage(trMsg);
      }
    } else if (msgName.equals("end-transition")) {
      TransitionMessage trMsg = new TransitionMessage(this);
      trMsg.isStart = false;
      trMsg.name = msg.getAttr("name");
      trMsg.fsm = msg.getAttr("fsm");
      trMsg.id = msg.getAttr("id");
      for (FsmListener lst : getListeners()) {
        lst.processTransitionMessage(trMsg);
      }
    } else if (msgName.equals("iteration-variable")) {
      IterVarMessage ivMsg = new IterVarMessage(this);
      ivMsg.type = msg.getAttr("type");
      ivMsg.name = msg.getAttr("name");
      ivMsg.value = msg.getAttr("value");
      ivMsg.fsm = msg.getAttr("fsm");
      for (FsmListener lst : getListeners()) {
        lst.processIterationVariable(ivMsg);
      }
    } else if (msgName.equals("summary")) {
      FsmSummaryMessage fsMsg = new FsmSummaryMessage(this);
      Map<String, String> data = new LinkedHashMap<String, String>();
      for (TraceNode tr : msg.getChildren()) {
        Map<String, String> attrs = tr.getAttrs();
        String name = attrs.get("name");
        String value = attrs.get("value");
        if (!"property".equals(tr.getName())
            || name == null
            || value == null
            || data.put(name, value) != null) {
          throw new IllegalArgumentException("Invalid summary message");
        }
      }
      fsMsg.data = Collections.unmodifiableMap(data);
      for (FsmListener lst : getListeners()) {
        lst.processSummaryMessage(fsMsg);
      }
    } else if (msgName.equals("browser-message")) {
      FsmBrowserMessage fbMsg = new FsmBrowserMessage(this);
      fbMsg.fsm = msg.getAttr("fsm");
      try {
        fbMsg.level = MessageLevel.valueOf(msg.getAttr("level"));
        fbMsg.msg = msg.getChildren().get(0).getText();
      } catch (Exception ex) {
        //for fixing bug 1879: instead of exception throwing only error message
        //throw new IllegalArgumentException("Invalid browser message", e);
        System.err.println("Invalid browser message");
      }
      for (FsmListener lst : getListeners()) {
        lst.processBrowserMessage(fbMsg);
      }
    } else {
      throw new IllegalArgumentException("Unsupported fsm node: " + msgName);
    }
  }

  @Override
  protected List<FsmListener> getListeners() {
    return (List<FsmListener>) super.getListeners();
  }
}
