/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.html;

import com.unitesk.aspectrace.ReportEntity;

import java.io.IOException;

/**
 * Report file.
 *
 * @author Sergey Groshev
 */
public abstract class ReportPage extends ReportEntity {

  public static final String DEFAULT_CHARSET = "utf-8";

  private ReportGenerator generator;
  private String path;
  private String charset;

  public ReportPage() {}

  public ReportPage(String path) {
    setPath(path);
  }

  // ------------------------------------------------------
  // Document path.

  /**
   * Sets a path to a page relative to the report root. If path is set as {@code null}
   * then the page wouldn't be generated.
   */
  public void setPath(String newPath) {
    this.path = newPath;
  }

  /** Gets path to page relative to report root. */
  public String getPath() {
    return path;
  }

  public String getRelativePathFrom(String from) {
    return getRelativePath(from, getPath());
  }

  public String getRelativePathFrom(ReportPage from) {
    return getRelativePath(from.getPath(), getPath());
  }

  public String getRelativePathTo(String to) {
    return getRelativePath(getPath(), to);
  }

  public String getRelativePathTo(ReportPage to) {
    return getRelativePath(getPath(), to.getPath());
  }

  /** Gets a relative path.
   * @return Path in the form of "../../../a/b/c/d.html" */
  protected static String getRelativePath(String from, String to) {
    String[] fromPath = from.split("/");
    String[] destPath = to.split("/");
    int fpl = fromPath.length - 1;
    int dpl = destPath.length - 1;
    int match = 0;
    while (match < fpl && match < dpl && fromPath[match].equals(destPath[match])) {
      match++;
    }

    StringBuilder sb = new StringBuilder();
    for (int i = match; i < fpl; i++) {
      sb.append("../");
    }
    for (int i = match; i < dpl; i++) {
      sb.append(destPath[i]).append('/');
    }
    sb.append(destPath[dpl]);
    return sb.toString();
  }

  // ------------------------------------------------------
  // Generation.

  /**
   * Generates complete page text.
   *
   * @throws IOException Input-Output error.
   */
  public abstract void generate() throws IOException;

  /** Sets explicitly page coding. */
  public void setCharset(String charset) {
    this.charset = charset;
  }

  /**
   * Gets a page coding.
   * If the coding is not set then a {@link ReportGenerator generator} default coding is returned.
   */
  public String getCharset() {
    if (charset != null) {
      return charset;
    }
    if (getGenerator() != null) {
      return getGenerator().getDefaultCharset();
    }
    return DEFAULT_CHARSET;
  }

  /**
   * Builds a reference map. It's possible to include pages in one another as sections. Method
   * is called by a report generator right before a generation start
   * when there is full set of report pages and resolvers.
   */
  public void buildRefs() {}

  /** Gets generator to which this report page is adjusted. */
  public ReportGenerator getGenerator() {
    return generator;
  }

  protected void setGenerator(ReportGenerator generator) {
    this.generator = generator;
  }
}
