/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.html;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;

/**
 * Static resource file (e.g. CSS or JS).
 *
 * <p>Overrides methods {@link #hashCode()} and {@link #equals(Object)} so that different components
 * may add all necessary resources in their generator: copies would be deleted automatically
 * so there wouldn't be any repeated copying.
 *
 * @author Sergey Groshev
 */
public class StaticResource extends ReportPage {
  private String inPath;
  public static final String GIF_PACKAGE = "com/unitesk/aspectrace/templates/gifs/";

  /**
   * Constructor.
   * @param inPath Source path in filesystem or in classpath.
   * @param outPath Result path relative root generation.
   */
  public StaticResource(String inPath, String outPath) {
    super(outPath);
    this.inPath = inPath;
  }

  @Override
  public void generate() throws IOException {
    InputStream is = null;
    OutputStream os = null;
    try {
      File out = new File(getGenerator().getGenerationDir(), getPath());
      File outDir = out.getParentFile();
      if (outDir != null && !outDir.isDirectory()) {
        outDir.mkdirs(); // Ignore errors here
      }
      final ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
      URL url = classLoader.getResource("com/unitesk/aspectrace/report.css");
      is = url.openStream();
      os = new BufferedOutputStream(new FileOutputStream(out));
      copy(is, os);
      is.close();
      os.close();
      // ---copying other files
      String[] files = {
        "bg.gif",
        "left.gif",
        "right.gif",
        "left_on.gif",
        "right_on.gif",
        "empty.gif",
        "failure.gif",
        "package.gif",
        "scenario.gif",
        "specification.gif",
        "method.gif"
      };
      for (int i = 0; i < files.length; i++) {
        String path = "/" + GIF_PACKAGE + files[i];
        is = getClass().getResourceAsStream(path);
        os = new FileOutputStream(new File(outDir, files[i]));
        copy(is, os);
        is.close();
        os.close();
      }
      // ---
    } finally {
      if (os != null) {
        os.close();
      }
      if (is != null) {
        is.close();
      }
    }
  }

  /** Copy entire contents from one stream to another. Does not close streams. */
  protected static void copy(InputStream is, OutputStream os) throws IOException {
    if (is == null) {
      System.err.println("?");
    }
    byte[] buffer = new byte[4096];
    for (; ; ) {
      int count = is.read(buffer);
      if (count < 0) {
        break;
      }
      os.write(buffer, 0, count);
    }
  }

  // ------------------------------------------------------
  // Same resources must be identified.

  @Override
  public int hashCode() {
    return 31 * inPath.hashCode() + getPath().hashCode();
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }
    StaticResource other = (StaticResource) obj;
    if (!inPath.equals(other.inPath)) {
      return false;
    }
    if (!getPath().equals(other.getPath())) {
      return false;
    }
    return true;
  }

  @Override
  public String toString() {
    return "StaticResource: " + getPath() + " (<= " + inPath + ')';
  }
}
