/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.impl;

import com.unitesk.aspectrace.Parameter;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceNode;
import com.unitesk.aspectrace.Tracer;

/**
 * Tracer for aspect "impl".
 *
 * @author Sergey Groshev
 */
public class ImplTracer {

  public static final String ASPECT = "impl";

  private static int callCounter = 0;

  private static void traceNode(TraceNode tn) {
    TraceMessage msg = new TraceMessage(ASPECT, tn);
    Tracer.traceMessage(msg);
  }

  private static String idToTag(String signature, int callId) {
    return signature + ':' + callId;
  }

  public static void traceMethodStart(String signature, Parameter... params) {

    Tracer.traceTagStart(ASPECT, signature, null);

    TraceNode tn = new TraceNode("start", "name", signature);
    for (Parameter p : params) {
      tn.addChild(p.toNode());
    }

    traceNode(tn);
  }

  /**.
   *  @return callId for use in {@link #traceMethodEndEx}. */
  public static int traceMethodStartEx(String signature, Parameter... params) {
    int res;
    synchronized (ImplTracer.class) {
      res = ++callCounter;
    }
    String tag = idToTag(signature, res);

    Tracer.traceTagStart(ASPECT, signature, null);
    Tracer.traceTagStart(ASPECT, tag, signature);

    TraceNode tn = new TraceNode("start", "name", signature, "id", Integer.toString(res));
    for (Parameter p : params) {
      tn.addChild(p.toNode());
    }
    traceNode(tn);

    return res;
  }

  public static void traceMethodEnd(String signature, Object retVal, Object exception) {
    TraceNode tn = new TraceNode("end", "name", signature);
    if (exception != null) {
      TraceNode tn2 = new TraceNode("exception", "type", exception.getClass().getName());
      tn2.addTextChild(exception.toString());
      tn.addChild(tn2);
    }
    if (retVal != null) {
      tn.addChild(TraceNode.createNodeWithText("return", retVal.toString()));
    }

    traceNode(tn);
    Tracer.traceTagEnd(ASPECT, signature, null);
  }

  public static void traceMethodEndEx(
      String signature, int callId, Object retVal, Object exception) {
    TraceNode tn = new TraceNode("end", "name", signature, "id", Integer.toString(callId));
    if (exception != null) {
      TraceNode tn2 = new TraceNode("exception", "type", exception.getClass().getName());
      tn2.addTextChild(exception.toString());
      tn.addChild(tn2);
    }
    if (retVal != null) {
      tn.addChild(TraceNode.createNodeWithText("return", retVal.toString()));
    }
    traceNode(tn);
    Tracer.traceTagEnd(ASPECT, idToTag(signature, callId), signature);
    Tracer.traceTagEnd(ASPECT, signature, null);
  }
}
