/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.provider;

import com.unitesk.aspectrace.AbstractTraceMessage;
import com.unitesk.aspectrace.MessageStream;
import com.unitesk.aspectrace.MessageStreamListener;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.TraceMessage;
import com.unitesk.aspectrace.TraceProcessorManager;
import com.unitesk.aspectrace.TraceProvider;

import java.io.IOException;

/**
 * Преобразователь из {@link MessageStream потока сообщений} в {@link TraceProvider источник
 * трассы}.
 *
 * @author Sergey Groshev
 */
public class StreamTraceProvider extends TraceProvider implements MessageStreamListener {
  protected String streamName;
  protected MessageStream stream;
  protected TraceProcessorManager tpm;

  public StreamTraceProvider(String streamName, TraceProcessorManager tpm) {
    this(null, streamName, tpm);
  }

  /**
   * Основной конструктор.
   *
   * @param stream Поток сообщений. Может быть {@code null}, в этом случае надо позже вызвать метод
   *     {@link #setStream(MessageStream)} до начала трассы.
   * @param streamName Имя потока. Используется для извещения обработчиков.
   * @param tpm Менеджер обработки.
   */
  public StreamTraceProvider(MessageStream stream, String streamName, TraceProcessorManager tpm) {
    this.streamName = streamName;
    this.tpm = tpm;
    if (stream != null) {
      setStream(stream);
    }
    if (tpm != null) {
      tpm.addProvider(this);
    }
  }

  /**
   * Установить поток сообщений.
   *
   * @param stream Поток. Может быть {@code null}, но в этом случае ничего не меняется.
   * @throws IllegalStateException Поток уже установлен.
   */
  public void setStream(MessageStream stream) throws IllegalStateException {
    if (this.stream != null) {
      throw new IllegalStateException("Stream already set");
    }
    if (stream != null) {
      this.stream = stream;
      stream.addListener(this);
    }
  }

  @Override
  public void handleStreamStart() {
    attachToManager(tpm);
    fireTraceStart(streamName);
  }

  @Override
  public void handleStreamMessage(AbstractTraceMessage msg) {
    switch (msg.getKind()) {
      case TRACE_MESSAGE:
        fireMessage((TraceMessage) msg);
        break;
      case TAG_MESSAGE:
        fireTagMessage((TagMessage) msg);
        break;
      case DEBUG_MESSAGE:
        break;
      default:
        throw new IllegalArgumentException("Unsupported message class: " + msg.getClass());
    }
  }

  @Override
  public void handleStreamEnd(boolean normal) {
    fireTraceEnd(streamName, normal);
  }

  /**
   * Запускает обработку сообщений в целевом PULL-потоке. Если целевой поток работает по модели
   * PUSH, то данный метод вызывать не следует.
   */
  @Override
  public void process() throws IOException, IllegalStateException, IllegalArgumentException {
    stream.process();
  }
}
