/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.ref;

import com.unitesk.aspectrace.ReportEntity;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Реестр компонентов-поставщиков контекстной информации. Поддерживает список зарегистрированных
 * поставщиков и обеспечивает доступ к ним компонентов, нуждающихся в контекстной информации.
 *
 * @author Sergey Groshev
 * @param <T> тип элементов отчёта
 */
public class ContextProviderRegistry<T extends ReportEntity> {

  /** Маркер, обозначающий запрос ко всем зарегистрированным поставщикам, независимо от их имён. */
  public static final String[] ALL = new String[0];

  protected Map<String, ContextProvider<T>> providers =
      new LinkedHashMap<String, ContextProvider<T>>();

  public ContextProviderRegistry() {}

  public void registerProvider(String name, ContextProvider<T> provider) {
    providers.put(name, provider);
  }

  public void unregisterProvider(String name) {
    providers.remove(name);
  }

  public ContextProvider<T> getProvider(String name) {
    return providers.get(name);
  }

  public ContextProvider<T>[] getAllProviders() {
    Collection<ContextProvider<T>> provs = providers.values();
    return provs.toArray(new ContextProvider[provs.size()]);
  }

  public String[] getNames() {
    Collection<String> names = providers.keySet();
    return names.toArray(new String[names.size()]);
  }

  /**
   * Получить поставщика-посредника, собирающего контекстную информацию от группы зарегистрированных
   * поставщиков.
   *
   * @param providerNames Имена интересующих поставщиков контекстной. информации. Поставщики с
   *     такими именами не обязаны присутствовать в реестре ни на момент вызова данного метода, ни
   *     на момент опроса полученного поставщика. Значение {@link #ALL} обозначает сбор информации
   *     ото всех провайдеров в том порядке, в котором они зарегистрированы.
   * @param targetClass <i>(Особенности Java)</i> необходимо указать тот класс, которым
   *     параметризован данный реестр. Сгенерированный поставщик будет возвращать массивы объектов
   *     этого класса.
   * @return Постащик, собирающий с указанных поставщиков из реестра контекстную информацию и
   *     возвращающий её единым списком. Порядок возвращаемых по запросу элементов отчёта
   *     соответствует указанному порядку поставщиков, а также порядку возвращаемых ими контекстных
   *     элементов.
   */
  public ContextProvider<T> getProxyProvider(String[] providerNames, final Class<T> targetClass) {
    final String[] names = providerNames == ALL ? ALL : providerNames.clone();
    return new ContextProvider<T>() {
      @Override
      public T[] getRelated(String channel, T target) {
        ArrayList<T> result = new ArrayList<T>();
        String[] nms = names == ALL ? getNames() : names;
        for (int i = 0; i < nms.length; i++) {
          ContextProvider<T> cp = getProvider(nms[i]);
          if (cp == null) {
            continue;
          }
          T[] rel = cp.getRelated(channel, target);
          if (rel == null) {
            continue;
          }
          for (T t : rel) {
            result.add(t);
          }
        }
        if (result.isEmpty()) {
          return null;
        }

        T[] array = (T[]) Array.newInstance(targetClass, result.size());
        return result.toArray(array);
      }
    };
  }
}
