/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.ref;

import com.unitesk.aspectrace.ReportEntity;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Управление разрешением ссылок между элементами отчета. Ссылки квалифицируются аспектом. Внутри
 * аспекта ресолверы вызываются в том порядке, в каком они добавлялись методом {@link
 * #addResolver(String, ReferenceResolver)}
 *
 * @author Sergey Groshev
 */
public class ReferenceManager<T extends ReportEntity> {

  protected Map<String, List<ReferenceResolver<T>>> resolvers =
      new HashMap<String, List<ReferenceResolver<T>>>();

  public void addResolver(String aspect, ReferenceResolver<T> resolver) {
    List<ReferenceResolver<T>> rsls = resolvers.get(aspect);
    if (rsls == null) {
      rsls = new ArrayList<ReferenceResolver<T>>();
      resolvers.put(aspect, rsls);
    }
    rsls.add(resolver);
  }

  public void removeResolver(String aspect, ReferenceResolver<T> resolver) {
    List<ReferenceResolver<T>> rsls = resolvers.get(aspect);
    if (rsls == null) {
      return;
    }
    rsls.remove(resolver);
    if (rsls.isEmpty()) {
      resolvers.remove(aspect);
    }
  }

  /**
   * Разрешение ссылки.
   *
   * @param aspect Аспект.
   * @param channel Привязка к каналу ({@code null}, если канал неважен)
   * @param ref Текстовая ссылка.
   * @return <b>Первый</b> удачный результат поиска.
   */
  public T resolve(String aspect, String channel, String ref) {
    List<ReferenceResolver<T>> rsls = resolvers.get(aspect);
    if (rsls == null) {
      return null;
    }
    for (ReferenceResolver<T> r : rsls) {
      T re = r.resolve(channel, ref);
      if (re != null) {
        return re;
      }
    }
    return null;
  }

  /**
   * Разрешение ссылки.
   *
   * @param aspect Аспект.
   * @param channel Привязка к каналу ({@code null}, если канал неважен)
   * @param ref Текстовая ссылка.
   * @return <b>Все</b> удачные результаты поиска в порядке, соответствующем порядку регистрации
   *     ресолверов.
   */
  public List<T> resolveMulti(String aspect, String channel, String ref) {
    List<ReferenceResolver<T>> rsls = resolvers.get(aspect);
    if (rsls == null) {
      return Collections.EMPTY_LIST;
    }
    ArrayList<T> res = new ArrayList<T>();
    for (ReferenceResolver<T> r : rsls) {
      T re = r.resolve(channel, ref);
      if (re != null) {
        res.add(re);
      }
    }
    if (res.isEmpty()) {
      return Collections.EMPTY_LIST;
    } else {
      return res;
    }
  }
}
