/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.common.ChannelMessage;
import com.unitesk.aspectrace.common.CommonListener;
import com.unitesk.aspectrace.error.ErrorListener;
import com.unitesk.aspectrace.error.FailureMessage;
import com.unitesk.aspectrace.fsm.FsmBrowserMessage;
import com.unitesk.aspectrace.fsm.FsmEndMessage;
import com.unitesk.aspectrace.fsm.FsmListener;
import com.unitesk.aspectrace.fsm.FsmStartMessage;
import com.unitesk.aspectrace.fsm.FsmStateMessage;
import com.unitesk.aspectrace.fsm.FsmSummaryMessage;
import com.unitesk.aspectrace.fsm.IterVarMessage;
import com.unitesk.aspectrace.fsm.TransitionMessage;
import com.unitesk.aspectrace.html.HeaderContext;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.ref.ContextProvider;
import com.unitesk.aspectrace.ref.ReferenceResolver;
import com.unitesk.aspectrace.templates.DocumentContext;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * Поточный генератор страниц, описывающих обходы сценарных КА. Создаёт и присоединяет к генератору
 * страницы сам, как только обнаруживает начало обхода КА. Обеспечивает расстановку тэгов,
 * разрешение ссылок и контекстные секции.
 */
public class FsmGraphProcessor
    implements FsmListener,
        ErrorListener,
        CommonListener,
        ConfigurableGenerator,
        ReferenceResolver<HtmlReportPage>,
        ContextProvider<HtmlReportPage> {
  public File reportDir;
  public static final String FSM_PREFIX = "fsm-";
  public static final String COMMON_TRANSITION_PREFIX = "transition-";

  private String previousMsgId = "";
  private String previousMsgName = "";
  private String previousMsgFsm = "";

  protected ReportGenerator rg;
  protected ErrorCollector ec;
  protected List<Integer> savedFailures = new ArrayList<Integer>();
  protected int curErrIdx = 0;

  private List<ContextProvider<HtmlReportPage>> providers =
      new ArrayList<ContextProvider<HtmlReportPage>>();

  protected boolean configTransitionTags = true;
  protected boolean configTrackFailures = true;
  protected boolean configDetailedIds = false;

  /**
   * Create new graph processor.
   *
   * @param rg Report generator for streamable page generation
   * @param ec Error collector to generate references to errors in transition (may be null).
   */
  public FsmGraphProcessor(ReportGenerator rg, ErrorCollector ec) {
    this.ec = ec;
    this.rg = rg;
  }

  // --------------- ConfigurableGenerator

  @Override
  public void configure(Properties config) {
    configTransitionTags =
        Boolean.parseBoolean(
            config.getProperty("FsmGraphProcessor.generate.transition.tags", "true"));
    configTrackFailures =
        Boolean.parseBoolean(config.getProperty("FsmGraphProcessor.track.failures", "true"));
    configDetailedIds =
        Boolean.parseBoolean(
            config.getProperty("FsmGraphProcessor.generate.detailed.ids", "false"));
  }

  /**
   * Этот метод вызывать обычно не надо, потому что все страницы добавляются в генератор немедленно
   * после создания для поточной генерации.
   */
  @Override
  public ReportPage[] getPages() {
    return allPages.toArray(new HtmlReportPage[allPages.size()]);
  }

  public void addSectionProvider(ContextProvider<HtmlReportPage> sp) {
    providers.add(sp);
  }

  // --------------- HtmlReportPage

  /** PrintWriter for current page. */
  protected Map<String, FsmPage> curPage = new HashMap<String, FsmPage>();

  protected Set<FsmPage> allPages = new LinkedHashSet<FsmPage>();

  public class FsmPage extends HtmlReportPage.Streamable {
    protected String curState = null;
    protected Transition curTrans = null;
    protected int curTransIdx = 0;
    protected String fsmName;
    protected String fsmAnchorName;
    protected String transitionPrefix;
    protected String traceDesc;

    public FsmPage(int index, String channelId, String fsmName) {
      super("scenario/fsm-report-" + index + ".html");
      setTitle("FSM traverse history: " + fsmName + " in channel " + channelId);
      setStyle(MainApp.CSS);
      rg.addResource(MainApp.CSS_SRC, MainApp.CSS);

      setAssociatedChannel(channelId);
      traceDesc = currentTraceDesc;

      this.fsmName = fsmName;
      fsmAnchorName = FSM_PREFIX + fsmName + '-' + fsmIndex;
      transitionPrefix = COMMON_TRANSITION_PREFIX + fsmName + '-' + fsmIndex + '-';
    }

    @Override
    public String getAspect() {
      return "fsm";
    }

    @Override
    public void startBody(PrintWriter pw) {
      // Сохраним pw, потому что поточной генерацией занимается в
      // основном обработчик, а не страница.

      // ---

      pw.println("<h1>" + getTitle() + "</h1>");
      pw.println("<p>Generated: " + Calendar.getInstance().getTime() + "</p>");
      DocumentContext doc = new DocumentContext();
      try {
        Properties props = new Properties();
        props.put(RuntimeConstants.RESOURCE_LOADER, "classpath");
        props.put("classpath.resource.loader.class", ClasspathResourceLoader.class.getName());
        Velocity.init(props);
        VelocityContext vc = new VelocityContext();
        String templateName = "overview-failure.vm";
        Template tmplt = Velocity.getTemplate(doc.getTemplatePackage() + templateName);
        HeaderContext docCont = new HeaderContext(getGenerator(), this);
        SummaryCollector sc = new SummaryCollector();

        vc.put("doc", doc);
        vc.put("current", this);
        vc.put("docContext", docCont);
        vc.put("summary", sc);

        tmplt.merge(vc, pw);

        ReportGenerator rg = new ReportGenerator();
        reportDir = rg.getGenerationDir();

      } catch (Exception ex) {
        ex.printStackTrace();
      }

      this.pw = pw;

      pw.println("<h1>FSM traverse history: " + fsmName + "</h1>");

      boolean first = true;
      for (ContextProvider<HtmlReportPage> sp : providers) {
        HtmlReportPage[] sections = sp.getRelated(getAssociatedChannel(), this);
        if (sections == null) {
          continue;
        }
        for (HtmlReportPage section : sections) {
          if (first) {
            first = false;
            pw.println("<h2>Context:</h2>");
          }
          section.setParent(this, null);
          section.generatePageBody(pw);
        }
      }
      if (!first) {
        pw.println("<br>");
      }

      pw.println(
          "<a class=\"transition\" name=\""
              + fsmAnchorName
              + "\">"
              + fsmAnchorName
              + ": <b>Start</b></a><br>");
    }

    @Override
    public String generateRef(HtmlReportPage from, String mark) {
      if (mark != null) {
        if (mark.equals(fsmAnchorName)) { // We have only one FSM per
          // page => mark may be
          // skipped
          return generateTextRef(
              from,
              null,
              configDetailedIds
                  ? "<b>FSM</b> "
                      + fsmName
                      + " in trace "
                      + traceDesc
                      + ", channel "
                      + getAssociatedChannel()
                  : "<b>FSM</b> " + fsmName);
        }
        if (mark.startsWith(COMMON_TRANSITION_PREFIX)) {
          if (mark.startsWith(transitionPrefix)) {
            mark = mark.substring(transitionPrefix.length());
          } else {
            mark = mark.substring(COMMON_TRANSITION_PREFIX.length());
          }
          return generateTextRef(
              from,
              mark,
              configDetailedIds
                  ? "<b>FSM</b> "
                      + fsmName
                      + " in trace "
                      + traceDesc
                      + ", channel "
                      + getAssociatedChannel()
                      + ", <b>transition</b> "
                      + mark
                  : "<b>FSM</b> " + fsmName + ", <b>transition</b> " + mark);
        }
      }
      return super.generateRef(from, mark);
    }

    PrintWriter getPw() {
      return pw;
    }

    public int getLevel() {
      return 1;
    }
  }

  // --------------- FsmListener

  int fsmIndex = 0;
  String currentTraceDesc;

  @Override
  public void processFsmStart(FsmStartMessage msg) {
    FsmPage page = new FsmPage(fsmIndex, msg.getChannel(), msg.getFullName());
    curPage.put(msg.getChannel(), page);
    page.setTagMarks(msg.getAllTags());
    allPages.add(page);
    rg.addPage(page); // will call startBody() and set field pw

    msg.getAdapter().addProcessorTag(page.fsmAnchorName, null);

    fsmIndex++;
  }

  @Override
  public void processFsmEnd(FsmEndMessage msg) {
    FsmPage page = curPage.get(msg.getChannel());
    processFailures(page);
    msg.getAdapter().removeProcessorTag(page.fsmAnchorName, null);

    curPage.put(msg.getChannel(), null);
  }

  @Override
  public void processStateMessage(FsmStateMessage msg) {
    if (msg.getName().equals(previousMsgName)
        && msg.getId().equals(previousMsgId)
        && msg.getFsm().equals(previousMsgFsm)) {
      return;
    }
    FsmPage page = curPage.get(msg.getChannel());
    if (page.curTrans != null) {
      throw new IllegalArgumentException("State inside transition");
    }

    page.curState = msg.getName();
    page.getPw().println("<b>State:</b> " + page.curState + "<br>");
    previousMsgName = msg.getName();
    previousMsgId = msg.getId();
    previousMsgFsm = msg.getFsm();
  }

  @Override
  public void processTransitionMessage(TransitionMessage msg) {
    FsmPage page = curPage.get(msg.getChannel());
    if (msg.isStart()) {
      page.curTransIdx++;
      page.curTrans = new Transition(page);
      page.curTrans.name = msg.getName();
      page.curTrans.idx = page.curTransIdx;
      if (configTransitionTags) {
        msg.getAdapter().addProcessorTag(page.curTrans.getFullMark(), page.fsmAnchorName);
      }
    } else {
      /* Streamable generation goes here */
      page.curTrans.dump(page.getPw());
      processFailures(page);
      if (configTransitionTags) {
        msg.getAdapter().removeProcessorTag(page.curTrans.getFullMark(), page.fsmAnchorName);
      }
      page.curTrans = null;
    }
  }

  private void processFailures(FsmPage page) {
    for (Integer i : savedFailures) {
      int idx = i.intValue();
      ErrorCollector.FailurePage errorPage = null;
      if (ec != null) {
        try {
          errorPage = ec.getPage(idx);
        } catch (Exception ex) {
          //TODO!
        }
      }
      if (errorPage == null) {
        page.getPw().print("<font color=\"#DD0000\">Error #" + (idx + 1) + " (no page)</font>");
      } else { // Add reference to error
        page.getPw().print(errorPage.generateRef(page, null));
      }
      page.getPw().println("<br>");
    }
    savedFailures.clear();
  }

  @Override
  public void processIterationVariable(IterVarMessage msg) {
    FsmPage page = curPage.get(msg.getChannel());
    if (page.curTrans == null) {
      throw new IllegalArgumentException("Iteration variable outside transition");
    }
    page.curTrans.itVars.add(msg.getType() + " " + msg.getName() + " = " + msg.getValue());
  }

  @Override
  public void processSummaryMessage(FsmSummaryMessage msg) {
    FsmPage page = curPage.get(msg.getChannel());
    page.getPw().println("<br><table class=\"section\" width=\"1005\"><tr><th>Summary");
    for (Map.Entry<String, String> me : msg.getData().entrySet()) {
      page.getPw().println("<tr><td>" + me.getKey() + " = " + me.getValue());
    }
    page.getPw().println("</table><br>");
  }

  @Override
  public void processBrowserMessage(FsmBrowserMessage msg) {
    FsmPage page = curPage.get(msg.getChannel());
    page.getPw()
        .println(
            "<font color=\"#00DD00\">Browser message: "
                + msg.getMsg()
                + " (severity: "
                + msg.getLevel()
                + ")</font><br>");
  }

  protected static class Transition {
    String name;
    int idx;
    ArrayList<String> itVars = new ArrayList<String>();
    FsmPage page;

    public Transition(FsmPage page) {
      this.page = page;
    }

    // Generate HTML presentation
    public void dump(PrintWriter pw) {
      String anchor = getAnchor();
      pw.println(
          "<br><a class=\"transition\" name=\""
              + anchor
              + "\"/>Transition "
              + anchor
              + ":</a><br>");
      pw.println("<b>Scenario method:</b> " + this.toString() + "<br>");
    }

    @Override
    public String toString() {
      if (itVars.isEmpty()) {
        return name;
      }

      StringBuilder sb = new StringBuilder();
      sb.append(name).append("( ");
      for (int i = 0; i < itVars.size(); i++) {
        if (i > 0) {
          sb.append(", ");
        }
        sb.append(itVars.get(i));
      }
      sb.append(" )");
      return sb.toString();
    }

    public String getAnchor() {
      return Integer.toString(idx);
    }

    public String getFullMark() {
      return page.transitionPrefix + getAnchor();
    }
  }

  // --------------- ErrorListener
  @Override
  public void processFailure(FailureMessage msg) {
    if (configTrackFailures) {
      // Save only indices, because ErrorCollector is not guaranteed
      // to be notified first.
      savedFailures.add(new Integer(curErrIdx));
      curErrIdx++;
    }
  }

  @Override
  public void processTagChange(TagMessage msg) {}

  @Override
  public void attachedToAdapter(String aspect, AspectAdapter adapter) {}

  // ------------------------------------------------
  // CommonListener

  @Override
  public void processingStarted(String traceDesc) {
    currentTraceDesc = traceDesc;
  }

  @Override
  public void processingFinished(String traceDesc, boolean normally) {
    currentTraceDesc = null;
  }

  @Override
  public void processChannelMessage(ChannelMessage msg) {}

  @Override
  public void processGlobalTagChange(TagMessage msg) {}

  // --------------- ReferenceResolver

  /**
   * Resolve something to FSM traverse page. Page will manage anchors by itself (see {@link
   * FsmPage#generateRef(HtmlReportPage, String)})
   */
  @Override
  public HtmlReportPage resolve(String channel, String ref) {
    for (FsmPage p : allPages) {
      String pgChan = p.getAssociatedChannel();
      if (pgChan == null || channel == null || pgChan.equals(channel)) {
        if (ref.equals(p.fsmAnchorName)) {
          return p;
        }
        if (ref.startsWith(p.transitionPrefix)) {
          return p;
        }
      }
    }
    return null;
  }

  // --------------- ContextProvider

  // Memory conservation note:
  // Let section be inner class, so section owner indirectly holds reference
  // to FsmGraphProcessor. There will be no memory leaks: all unneeded
  // information will be discarded.
  // Also keep reference to current transition (otherwise it will be replaced
  // by newer transitions).
  @Override
  public HtmlReportPage[] getRelated(String channel, final HtmlReportPage target) {
    FsmPage page = curPage.get(channel);
    if (page == null) {
      return null;
    }

    final String state = page.curState != null ? page.curState : "<i>undefined</i>";
    final Transition trans = page.curTrans;
    final FsmPage fsmPage = page;
    HtmlReportPage hrp =
        new HtmlReportPage() {
          @Override
          public void generatePageBody(PrintWriter pw) {
            pw.println("<table class=\"section\" width=\"100%\">");
            pw.println("<tr><th colspan=\"2\">FSM info");
            pw.println("<tr><td>Scenario<td>" + fsmPage.generateRef(target, fsmPage.fsmAnchorName));
            pw.println("<tr><td>State<td>" + state);
            if (trans != null) {
              pw.println(
                  "<tr><td>Transition<td>"
                      + fsmPage.generateRef(target, trans.getFullMark())
                      + ": <b>"
                      + trans
                      + "</b>");
            }
            pw.println("</table>");
          }
        };
    return new HtmlReportPage[] {hrp};
  }
}
