/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.TraceProcessorManager;
import com.unitesk.aspectrace.TraceProvider;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.provider.XmlFileParser;
import com.unitesk.aspectrace.ref.ContextProviderRegistry;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

public class GeneralGenerateReportCTESK {

  static final String DEFAULT_PROPERTIES = "GeneratorCTESK.properties";

  public static final String CSS = "report.css";
  public static final String CSS_SRC = "/com/unitesk/aspectrace/sampleApp/report.css";

  private TraceProcessorManager tpm = new TraceProcessorManager();
  private ReportGenerator rg = new ReportGenerator();
  private ContextProviderRegistry<HtmlReportPage> registry =
      new ContextProviderRegistry<HtmlReportPage>();

  private CoverageCollectorCTESK cc;
  private ErrorCollector ec;
  private OracleCollector oc;
  private FsmGraphProcessor gp;
  private TestProcessor tp;
  private TraceTracker tt;
  private RequirementProcessor reqtp;

  protected static boolean tracingErrors = true;
  protected static boolean tracingCoverage = true;
  protected static boolean tracingFSM = true;
  protected static boolean tracingRequirements = false;

  private Properties config;
  private List<TraceProvider> traces;

  private static void usage() {
    System.err.println(
        "Usage: GenerateReportCTESK [-config config_file]"
            + " [-f] [-e] [-c] [-r] [-d report_dir] sources");
    System.err.println("  where source ::= trace_name | -tcp port_number");
    System.err.println("  default config file is " + DEFAULT_PROPERTIES);
  }

  private static InputStream openResourceOrFile(String resourceOrFile) throws IOException {
    URL resourceURL = GeneralGenerateReportCTESK.class.getResource("/" + resourceOrFile);
    if (resourceURL != null) {
      return GeneralGenerateReportCTESK.class.getClassLoader().getResourceAsStream(resourceOrFile);
    }
    return new FileInputStream(resourceOrFile);
  }

  public static void main(String[] args) {
    Properties config = new Properties();
    String configPath = DEFAULT_PROPERTIES;
    String outDir = null;
    List<TraceProvider> traces = new ArrayList<TraceProvider>();
    try {
      for (int i = 0; i < args.length; i++) {
        if (args[i].equals("-d")) {
          outDir = args[++i];
        } else if (args[i].equals("-e")) {
          tracingErrors = false;
        } else if (args[i].equals("-c")) {
          tracingCoverage = false;
        } else if (args[i].equals("-f")) {
          tracingFSM = false;
        } else if (args[i].equals("-r")) {
          tracingRequirements = false;
        } else {
          String[] tracesFromArgs = args[i].split(";");
          for (String str : tracesFromArgs) {
            traces.add(new XmlFileParser(str));
          }
        }
      }
    } catch (RuntimeException re) { // including IndexOutOfBoundsException
      usage();
      return;
    }

    SummaryCollector.setTracingRequirements(tracingRequirements);

    InputStream configStream = null;
    try {
      configStream = openResourceOrFile(configPath);
      config.load(configStream);
      configStream.close();
    } catch (Exception ex) {
      throw new RuntimeException("Failed to load properties from " + configPath, ex);
    }

    if (outDir == null) {
      outDir =
          config.getProperty("default.out", "Reports/Report") + "_" + System.currentTimeMillis();
    }

    GeneralGenerateReportCTESK ggrct = new GeneralGenerateReportCTESK(config, traces, outDir);
    ggrct.generateFromTrace();
  }

  public GeneralGenerateReportCTESK(
      Properties config, List<TraceProvider> traces, String generationDir) {
    this.config = config;
    this.traces = traces;

    rg.setDefaultCharset(config.getProperty("page.encoding", "utf-8"));

    File repPath = new File(generationDir);

    rg.setGenerationDir(repPath);

    try {
      initComponents();
    } catch (Exception ex) {
      ex.printStackTrace();
      throw new IllegalStateException(ex);
    }
  }

  /**
   * Создать компоненты нашего отчёта, присоединить к источнику трассы и сконфигурировать связи
   * между ними.
   *
   * @throws Exception Проблемы конфигурации обработчиков трассы
   */
  void initComponents() throws Exception {
    // Присоединить адаптер для проверки синтаксиса пустого аспекта

    // ----- Создать и зарегистрировать компоненты

    tt = new TraceTracker();
    tpm.addListener("", tt);
    registry.registerProvider("TraceTracker", tt);

    boolean useHgClr = Boolean.parseBoolean(config.getProperty("CoverageCollector.huge", "false"));
    cc = useHgClr ? new HugeCoverageCollectorCTESK(config) : new CoverageCollectorCTESK(config);

    // ---
    if (tracingCoverage) {
      tpm.addListener("coverage", cc);
      rg.addRefResolver("coverage", cc);
    }

    // ---
    if (tracingErrors) {
      ec = new ErrorCollector(config);
      tpm.addListener("error", ec);
    }
    // ---

    oc = new OracleCollector(config);
    tpm.addListener("oracle", oc);
    rg.addRefResolver("oracle", oc);
    registry.registerProvider("OracleCollector", oc);

    boolean useImpl =
        Boolean.parseBoolean(config.getProperty("FsmGraphProcessor.track.impl", "false"));
    useImpl = true;
    // Will attach pages to generator automatically
    if (tracingFSM) {
      if (tracingErrors) {
        gp =
            useImpl
                ? new FsmAndImpl(rg, ec)
                : new FsmGraphProcessor(
                    rg, ec); // Will behave differently ONLY if attached as impl listener
      } else {
        gp = new FsmGraphProcessor(rg, null);
      }
      gp.configure(config);
      tpm.addListener("fsm", gp);
      tpm.addListener("error", gp);
      tpm.addListener("", gp);
      if (useImpl) {
        tpm.addListener("impl", gp);
      }
      rg.addRefResolver("fsm", gp);
      registry.registerProvider("FsmGraphProcessor", gp);
    }

    tp = new TestProcessor(config);

    tpm.addListener("test", tp);
    rg.addRefResolver("test", tp);
    registry.registerProvider("TestProcessor", tp);

    if (tracingRequirements) {
      reqtp = new RequirementProcessor(config);
      tpm.addListener("requirement", reqtp);
      rg.addRefResolver("requirement", reqtp);
      registry.registerProvider("RequirementProcessor", reqtp);
    }

    // ----- Сконфигурировать связи между компонентами

    if (tracingErrors) {
      ec.addSectionProvider(
          registry.getProxyProvider(
              getListPropOrAll("ErrorCollector.section.providers"), HtmlReportPage.class));
    }

    if (tracingFSM) {
      gp.addSectionProvider(
          registry.getProxyProvider(
              getListPropOrAll("FsmGraphProcessor.section.providers"), HtmlReportPage.class));
    }

    if (tracingErrors) {
      setAutoRefs("error");
    }

    if (tracingFSM) {
      setAutoRefs("fsm");
    }

    if (tracingCoverage) {
      setAutoRefs("coverage");
    }

    setAutoRefs("test");

    if (tracingRequirements) {
      setAutoRefs("requirement");
    }
  }

  private String[] getListProp(String propName) {
    return config.getProperty(propName, "").split("\\s*,\\s*");
  }

  private String[] getListPropOrAll(String propName) {
    String propValue = config.getProperty(propName, "");
    if (propValue.equals("*")) {
      return ContextProviderRegistry.ALL;
    } else {
      return propValue.split("\\s*,\\s*");
    }
  }

  private void setAutoRefs(String targetAspect) {
    for (String asp : getListProp("autoref." + targetAspect)) {
      boolean biDirect;
      if (asp.endsWith("+")) {
        biDirect = true;
        asp = asp.substring(0, asp.length() - 1);
      } else {
        biDirect = false;
      }
      rg.addAutoRef(targetAspect, asp, biDirect);
    }
  }

  void generateFromTrace() {
    try {
      for (TraceProvider trace : traces) {
        tpm.addProvider(trace);
      }
      tpm.process();
    } catch (Exception ex) {
      ex.printStackTrace();
      System.err.println();
      System.err.println("Trying to generate report after failure...");
    }

    try {
      generateReport();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  void generateReport() throws IOException {
    System.out.println("Generating report...");

    HtmlReportPage hrp = rg.addIndexPage();
    hrp.setTitle("Report generated by GeneralGenerateReportCTESK");
    rg.addDetailsPage(hrp).setTitle("Details");
    rg.addGenericReport("FSM", "fsm", hrp, "fsmList");
    rg.addGenericReport("Failures", "error", hrp, "failuresList");
    rg.addGenericReport("Coverage", "coverage", hrp, "coveragesList");

    if (tracingRequirements) {
      HtmlReportPage allDB = rg.addAllDtBsPage();
    }

    rg.addPages(tp.getPages());

    if (tracingRequirements) {
      rg.addPages(reqtp.getPages());
    }

    if (tracingErrors) {
      rg.addPages(ec.getPages());
    }

    rg.addPages(cc.getPages());
    rg.addPages(oc.getPages());

    // А это не работает, потому что наша страница оракула не имеет чёткой
    // привязки к месту в трассе, а значит, не имеет контекста, за который
    // её можно было бы привязать к сценарию с помощью механизма ссылок
    // (однако, её можно привязать через механизм контекстной информации).
    // rg.addAutoRef("oracle", "fsm");

    // TODO: Сделать ссылки на вызов оракула только для покрытий аспекта impl
    // или для покрытий всех аспектов, кроме oracle
    // rg.addAutoRef("coverage", "oracle");

    rg.generateReport();

    System.out.println("Report generated successfully");
  }
}
