/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.coverage.CoverageMessage;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.html.StaticResource;
import com.unitesk.aspectrace.ref.ReferenceResolver;
import com.unitesk.coverage.CollectableCoverageStructure;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageStructure;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 * Собирает информацию о покрытиях. Если поддерево структуры покрытия имеет слишком много листьев и
 * притом больше одного потомка - разбивает его на несколько иерархически организованных страниц.
 *
 * <p>Для каждого достигнутого листового элемента запоминает количество попаданий.
 *
 * <p>Для каждой структуры покрытия генерирует страницу и разрешает ссылки на неё (через интерфейс
 * {@link ReferenceResolver}) с параметрами:<br>
 * <code>aspect="coverage"</code> и <code>ref=targetAspect + ":" + coverageName</code>
 *
 * <p>Использует часть кода обработчика {@link CoverageCollector суперкласса}, но пользуется
 * собственным классом страниц.
 *
 * @author Sergey Groshev
 */
public class HugeCoverageCollector extends CoverageCollector {
  public static final int DEFAULT_THRESHOLD = 100;
  public static final String PROP_THRESHOLD = "HugeCoverageCollector.threshold";

  public HugeCoverageCollector() {
    configFirstCovered = false;
    configHits = true;
  }

  public HugeCoverageCollector(Properties config) {
    configure(config);
  }

  protected int threshold = DEFAULT_THRESHOLD;

  @Override
  public void configure(Properties config) {
    super.configure(config);

    String str = config.getProperty(PROP_THRESHOLD);
    if (str != null) {
      try {
        threshold = Integer.parseInt(str);
      } catch (NumberFormatException nfe) {
        // TODO!
      }
    }
  }

  // ------------------------------------------------------
  // CoverageListener

  protected class ElemInfo extends HtmlReportPage {
    int hitCount = 0;

    public ElemInfo(CoverageMessage cm, CoverageElement ce) {
      if (configFirstCovered) { // For auto-referencing
        setTagMarks(cm.getAllTags());
      }
      setAssociatedChannel(firstCoveredInChannel.get(ce));
    }

    // For auto-referencing
    @Override
    public String getAspect() {
      return "coverage";
    }

    @Override
    public void generatePageBody(PrintWriter pw) {}
  }

  protected Map<CoverageElement, ElemInfo> elemInfos = new HashMap<CoverageElement, ElemInfo>();

  @Override
  public void processCoverageMessage(CoverageMessage cm) {
    CoverageElement ce = getCE(cm); // Не требует вызова normalize()
    ElemInfo hi = elemInfos.get(ce);
    if (hi == null) {
      hi = new ElemInfo(cm, ce);
      elemInfos.put(ce, hi);
    }
    hi.hitCount++;
  }

  // ------------------------------------------------------
  // Page generation

  private Map<CoverageElement, Integer> leafCounts = new HashMap<CoverageElement, Integer>();

  protected int getLeafCount(CoverageElement ce) {
    Integer ii = leafCounts.get(ce);
    if (ii != null) {
      return ii.intValue();
    }

    CoverageElement[] children = ce.getExtendedChildren();
    if (children.length == 0) {
      return 1;
    }
    int lf = 0;
    for (CoverageElement child : children) {
      lf += getLeafCount(child);
    }

    leafCounts.put(ce, new Integer(lf));
    return lf;
  }

  protected Map<CoverageElement, CoveragePage> pages = new HashMap<CoverageElement, CoveragePage>();

  @Override
  public ReportPage[] getPages() {
    // Стратегия заполнения списка страниц: вглубь
    List<ReportPage> pages = new ArrayList<ReportPage>();
    for (Map<String, CoverageStructure> s0 : structs.values()) {
      for (CoverageStructure str : s0.values()) {
        CoverageElement root = str.getRoot();
        int height = getLeafCount(root);
        CoveragePage rootPage = getPage(root);
        pages.add(rootPage);
        if (height >= threshold) {
          dumpSubPages(pages, root, rootPage);
        }
      }
    }

    if (pages.isEmpty()) {
      return new ReportPage[0];
    }
    ReportPage[] rp = pages.toArray(new ReportPage[pages.size() + 1]);
    rp[rp.length - 1] = new StaticResource(MainApp.CSS_SRC, MainApp.CSS);
    return rp;
  }

  protected void dumpSubPages(
      Collection<ReportPage> collector, CoverageElement elem, CoveragePage parentPage) {
    CoverageElement[] children = elem.getExtendedChildren();
    for (CoverageElement child : children) {
      ElemInfo elemi = getLeaf(child);
      if (elemi != null) {
        collector.add(elemi);
        elemi.setParent(parentPage, null);
      }

      int height = getLeafCount(child);
      if (height >= threshold) {
        if (children.length > 1) {
          CoveragePage myPage = getPage(child);
          myPage.setMasterPage(parentPage);
          collector.add(myPage);
          dumpSubPages(collector, child, myPage);
        } else { // Если данный потомок - единственный, то не заводим отдельную страницу
          dumpSubPages(collector, child, parentPage);
        }
      }
    }
  }

  /**
   * Найти страницу для корневого элемента покрытия. Если страницы нет - она создаётся и
   * возвращается.
   */
  protected CoveragePage getPage(CoverageElement ce) {
    CoveragePage cp = pages.get(ce);
    if (cp == null) {
      cp = new CoveragePage(ce);
      pages.put(ce, cp);
    }
    return cp;
  }

  protected ElemInfo getLeaf(CoverageElement ce) {
    return elemInfos.get(normalize(ce));
  }

  protected class CoveragePage extends HtmlReportPage {
    // extends CoverageCollector.CoveragePage

    protected int coveredCount;
    protected CoverageElement rootElem;

    protected CoveragePage(CoverageElement rootElem) {
      this.rootElem = rootElem;
      CoverageStructure cs = rootElem.getStructure();
      String[] namePath = rootElem.getNamePath();

      String path = cs.getTargetAspect() + '-' + cs.getName();
      if (rootElem.getRawParent() != null) {
        path = path + '-' + join(namePath, "-");
      }
      path = path.replaceAll("[<>:/\\\\]", "_");
      setPath("coverage/" + path + ".html");

      String title = "Coverage report: " + cs.getTargetAspect() + " :: " + cs.getName();
      if (rootElem.getRawParent() != null) {
        title = title + " / " + join(namePath, " / ");
      }
      setTitle(title);
      setStyle(MainApp.CSS);
    }

    @Override
    public String getAspect() {
      return "coverage";
    }

    @Override
    public void generatePageBody(PrintWriter pw) {
      coveredCount = 0;

      CoverageElement[] elements = rootElem.getExtendedChildren();
      CoverageStructure cs = rootElem.getStructure();
      int maxDepth = getMaxDepth(elements);
      int width = maxDepth + (configHits ? 1 : 0) + (configFirstCovered ? 1 : 0);

      // Страница: заголовок над таблицей
      if (getParentPage() == null) {
        pw.println("<h1>Coverage report</h1>");
      }

      pw.println("<table class=\"coverage\">");
      pw.println("<tr><td colspan=\"" + width + "\">");
      pw.println("<b>Aspect:</b> " + cs.getTargetAspect() + "<br>");
      pw.println("<b>Coverage name:</b> " + encodeSimple(cs.getName()));

      if (rootElem.getRawParent() != null) {
        CoverageElement[] path = rootElem.getPath();
        pw.println("<br>");
        pw.println("<b>Depth:</b> " + path.length + "<br>");
        pw.println("<b>Path:</b> ");

        // Add root, remove element itself
        CoverageElement[] followups = new CoverageElement[path.length];
        followups[0] = cs.getRoot();
        System.arraycopy(path, 0, followups, 1, path.length - 1);
        for (int i = 0; i < followups.length; i++) {
          CoverageElement ce = followups[i];
          if (i > 0) {
            pw.print(" &raquo; ");
          }
          CoveragePage uplink = pages.get(ce);
          if (uplink != null) {
            pw.print(
                uplink.generateTextRef(
                    this, null, ce.getRawParent() == null ? "(Root)" : ce.getName()));
          } else {
            pw.print(ce.getName()); // Always non-root
          }
        }
        pw.println(" &raquo; " + rootElem.getName());
      }

      if (configHits || configFirstCovered) {
        pw.println("<tr><th colspan=\"" + maxDepth + "\">Coverage elements");
      }
      if (configHits) {
        pw.println("<th style='text-align: center;'>Hits");
      }
      if (configFirstCovered) {
        pw.println("<th>First covered");
      }

      pw.println("<tr>");
      generateRows(pw, elements, maxDepth);

      pw.println("<tr><th colspan=\"" + width + "\">");
      pw.print("Total coverage: <b>" + coveredCount + " / ");
      if (cs instanceof CollectableCoverageStructure) {
        pw.print("<font color=\"#FF0000\">?</font></b>");
      } else {
        pw.print(getHeight(rootElem.getChildren()));
      }
      pw.println("</b>");

      pw.println("</table><br/>");
    }

    protected void generateRows(PrintWriter pw, CoverageElement[] elements, int colspan) {

      for (int i = 0; i < elements.length; i++) {
        CoverageElement ce = elements[i];

        boolean isCovered = getHits(normalize(ce)) != 0;
        String tdClass =
            !isCovered ? "uncovered" : hasUncoveredLeaves(ce) ? "semicovered" : "covered";

        if (pages.containsKey(ce)) {
          pw.print("<td class=\"" + tdClass + "\"");
          if (colspan > 1) {
            pw.print(" colspan=" + colspan);
          }
          pw.println("><i> &raquo;&raquo;&raquo; ");
          pw.println(getPage(ce).generateTextRef(this, null, ce.getName()));
          pw.println(" (" + getLeafCount(ce) + " leaves)");
          pw.println("</i></td>");

          if (configHits) {
            pw.print("<td class=\"" + tdClass + "\"align=center><i>" + getHits(ce) + "</i></td>");
          }
          if (configFirstCovered) {
            pw.print("<td class=\"" + tdClass + "\">&nbsp;");
            ElemInfo elemi = getLeaf(ce);
            if (elemi != null) {
              elemi.generateRefs(pw, true);
            } else {
              pw.print("&raquo;&raquo;&raquo;");
            }
          }
          pw.println("</tr>");
          if (isCovered) {
            coveredCount++;
          }
          continue;
        }

        CoverageElement[] children = ce.getExtendedChildren();
        int height = getHeight(children);

        String name = ce.getName();
        if (name == null || name.length() == 0) { // null for dummy child
          name = "&minus;";
        }

        if (i > 0) {
          pw.println("<tr>");
        }

        pw.print("<td class=\"" + tdClass + "\"");
        if (height > 1) {
          pw.print(" rowspan=" + height);
        }
        if (children.length == 0 && colspan > 1) {
          pw.print(" colspan=" + colspan);
        }
        pw.print(">");

        pw.print(name);
        pw.println("</td>");

        if (children.length == 0) {
          if (isCovered) {
            coveredCount++;
            if (configHits) {
              pw.print("<td class=\"" + tdClass + "\"align=center>" + getHits(ce));
            }
            if (configFirstCovered) {
              pw.print("<td class=\"" + tdClass + "\">&nbsp;");
              getLeaf(ce).generateRefs(pw, true);
            }
          } else {
            if (configHits) {
              pw.print("<td class=\"uncovered\" align=\"center\">0");
            }
            if (configFirstCovered) {
              pw.print("<td class=\"uncovered\">&nbsp;&minus;");
            }
          }
          pw.println("</tr>");
        } else {
          generateRows(pw, children, colspan - 1);
        }
      }
    }

    protected int getHits(CoverageElement ce) {
      ElemInfo elemi = elemInfos.get(normalize(ce));
      int sum = (elemi != null) ? elemi.hitCount : 0;
      for (CoverageElement child : ce.getChildren()) {
        sum += getHits(child);
      }
      return sum;
    }

    protected boolean hasUncoveredLeaves(CoverageElement ce) {
      if (ce.isLeaf()) {
        return !elemInfos.containsKey(normalize(ce));
      }

      CoverageElement[] children = ce.getExtendedChildren();
      for (int i = 0; i < children.length; i++) {
        if (hasUncoveredLeaves(children[i])) {
          return true;
        }
      }
      return false;
    }

    protected int getHeight(CoverageElement[] elems) {
      if (elems.length == 0) {
        return 1;
      }

      int sum = 0;
      for (int i = 0; i < elems.length; i++) {
        CoverageElement ce = elems[i];
        if (ce != rootElem && pages.containsKey(ce)) {
          sum++;
        } else {
          sum += getHeight(ce.getExtendedChildren());
        }
      }
      return sum;
    }

    protected int getMaxDepth(CoverageElement[] elems) {
      int maxDepth = 1;
      for (int i = 0; i < elems.length; i++) {
        CoverageElement[] children = elems[i].getChildren();
        if (children.length > 0) {
          int depth = getMaxDepth(children) + 1;
          if (maxDepth < depth) {
            maxDepth = depth;
          }
        }
      }
      return maxDepth;
    }
  } // end class CoveragePage

  protected static String join(String[] array, String separator) {
    StringBuilder sb = new StringBuilder();
    for (int i = 0; i < array.length; i++) {
      if (i > 0) {
        sb.append(separator);
      }
      sb.append(array[i]);
    }
    return sb.toString();
  }

  // ------------------------------------------------------
  // ReferenceResolver

  @Override
  protected HtmlReportPage resolve(CoverageStructure cs) {
    return getPage(cs.getRoot()); // will create page if needed
  }
}
