/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.MessageLevel;
import com.unitesk.aspectrace.Parameter;
import com.unitesk.aspectrace.TraceProcessorManager;
import com.unitesk.aspectrace.TraceProvider;
import com.unitesk.aspectrace.TraceUtils;
import com.unitesk.aspectrace.Tracer;
import com.unitesk.aspectrace.common.CommonTracer;
import com.unitesk.aspectrace.coverage.CoverageTracer;
import com.unitesk.aspectrace.fsm.FsmTracer;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportGenerator;
import com.unitesk.aspectrace.oracle.MarkedPath;
import com.unitesk.aspectrace.oracle.OracleTracer;
import com.unitesk.aspectrace.provider.XmlFileParser;
import com.unitesk.aspectrace.provider.XmlSocketServer;
import com.unitesk.aspectrace.ref.ContextProviderRegistry;
import com.unitesk.aspectrace.test.TestTracer;
import com.unitesk.coverage.CoverageElement;
import com.unitesk.coverage.CoverageStructure;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

/**
 * Пример использования библиотеки Aspectrace: генерирует отчёты из предложенных трасс, а также
 * может генерировать небольшой пример трассы.
 *
 * <p>Смотреть в первую очередь код методов:
 *
 * <ul>
 *   <li>{@link #generateTrace(String, boolean)} - генерировать пример трассы (в файл или в отчёт на
 *       лету, в зависимости от настроек в файле {@code MainApp.properties}).
 *   <li>{@link #initComponents()} - сконфигурировать генератор отчётов.
 *   <li>{@link #generateFromTrace()} - считать входные трассы, обработать их, сгенерировать отчёт.
 *   <li>{@link #generateReport()} - сгенерировать отчёт, когда все трассы уже обработаны.
 * </ul>
 */
public class MainApp {

  static final String DEFAULT_PROPERTIES = "./MainApp.properties";

  public static final String CSS = "report.css";
  public static final String CSS_SRC = "/com/unitesk/aspectrace/sampleApp/report.css";

  private TraceProcessorManager tpm = new TraceProcessorManager();
  private ReportGenerator rg = new ReportGenerator();
  private ContextProviderRegistry<HtmlReportPage> registry =
      new ContextProviderRegistry<HtmlReportPage>();

  private CoverageCollector cc;
  private ErrorCollector ec;
  private OracleCollector oc;
  private FsmGraphProcessor gp;
  private TestProcessor tp;
  private TraceTracker tt;

  private Properties config;
  private List<TraceProvider> traces;

  private static void usage() {
    System.err.println("Usage: MainApp [-config config_file] [-d report_dir]  -sample | sources");
    System.err.println("  where source ::= trace_name | -tcp port_number");
    System.err.println("  default config file is " + DEFAULT_PROPERTIES);
  }

  private static InputStream openResourceOrFile(String resourceOrFile) throws IOException {
    URL resourceURL = MainApp.class.getResource("/" + resourceOrFile);
    if (resourceURL != null) {
      return MainApp.class.getClassLoader().getResourceAsStream(resourceOrFile);
    }
    return new FileInputStream(resourceOrFile);
  }

  public static void main(String[] args) {
    Properties config = new Properties();
    String configPath = DEFAULT_PROPERTIES;
    String outDir = null;
    List<TraceProvider> traces = new ArrayList<TraceProvider>();
    boolean sample = false;

    try {
      for (int i = 0; i < args.length; i++) {
        if (args[i].equals("-d")) {
          outDir = args[++i];
        } else if (args[i].equals("-config")) {
          configPath = args[++i];
        } else if (args[i].equals("-sample")) {
          sample = true;
        } else if (args[i].equals("-tcp")) {
          int port = Integer.parseInt(args[++i]);
          XmlSocketServer sockServ = new XmlSocketServer(port, true);
          traces.add(sockServ);
        } else {
          traces.add(new XmlFileParser(args[i]));
        }
      }
    } catch (RuntimeException re) { // including IndexOutOfBoundsException
      usage();
      return;
    } catch (IOException ioe) {
      ioe.printStackTrace();
      return;
    }

    if (sample != traces.isEmpty()) { // sample && !traces.isEmpty() || !sample && traces.isEmpty()
      usage();
      return;
    }

    InputStream configStream = null;
    try {
      configStream = openResourceOrFile(configPath);
      config.load(configStream);
      configStream.close();
    } catch (Exception ex) {
      throw new RuntimeException("Failed to load properties from " + configPath, ex);
    }

    if (outDir == null) {
      outDir = config.getProperty("default.out", "!Report");
    }

    if (sample) {
      String trace = config.getProperty("default.trace");
      MainApp ma;

      if (trace != null) {
        // Use case 1: Генерировать трассу, затем из неё отчёт.
        ma =
            new MainApp(
                config,
                Collections.singletonList((TraceProvider) new XmlFileParser(trace)),
                outDir);
        ma.generateTrace(trace, false);
        ma.generateFromTrace();
      } else {
        // Use case 2: Генерировать отчёт на лету, без трассы.
        ma = new MainApp(config, Collections.EMPTY_LIST, outDir);
        ma.generateTrace(null, true);
      }
      /* Use case 3: Генерировать трассу и одновременно отчёт на лету
       * -- в данном примере не используется */

      return;
    }

    MainApp ma = new MainApp(config, traces, outDir);
    ma.generateFromTrace();
  }

  MainApp(Properties config, List<TraceProvider> traces, String generationDir) {
    this.config = config;
    this.traces = traces;

    rg.setDefaultCharset(config.getProperty("page.encoding", "utf-8"));
    rg.setGenerationDir(new File(generationDir));

    try {
      initComponents();
    } catch (Exception ex) {
      ex.printStackTrace();
      throw new IllegalStateException(ex);
    }
  }

  /**
   * Создать компоненты нашего отчёта, присоединить к источнику трассы и сконфигурировать связи
   * между ними.
   *
   * @throws Exception Проблемы конфигурации обработчиков трассы
   */
  void initComponents() throws Exception {
    // Присоединить адаптер для проверки синтаксиса пустого аспекта

    // ----- Создать и зарегистрировать компоненты

    tt = new TraceTracker();
    tpm.addListener("", tt);
    registry.registerProvider("TraceTracker", tt);

    boolean useHgClr = Boolean.parseBoolean(config.getProperty("CoverageCollector.huge", "false"));
    cc = useHgClr ? new HugeCoverageCollector(config) : new CoverageCollector(config);
    tpm.addListener("coverage", cc);
    rg.addRefResolver("coverage", cc);

    ec = new ErrorCollector(config);
    tpm.addListener("error", ec);

    oc = new OracleCollector(config);
    tpm.addListener("oracle", oc);
    rg.addRefResolver("oracle", oc);
    registry.registerProvider("OracleCollector", oc);

    boolean useImpl =
        Boolean.parseBoolean(config.getProperty("FsmGraphProcessor.track.impl", "false"));
    // Will attach pages to generator automatically
    gp =
        useImpl
            ? new FsmAndImpl(rg, ec)
            : new FsmGraphProcessor(
                rg, ec); // Will behave differently ONLY if attached as impl listener
    gp.configure(config);
    tpm.addListener("fsm", gp);
    tpm.addListener("error", gp);
    tpm.addListener("", gp);
    if (useImpl) {
      tpm.addListener("impl", gp);
    }
    rg.addRefResolver("fsm", gp);
    registry.registerProvider("FsmGraphProcessor", gp);

    tp = new TestProcessor(config);
    tpm.addListener("test", tp);
    rg.addRefResolver("test", tp);
    registry.registerProvider("TestProcessor", tp);

    // ----- Сконфигурировать связи между компонентами

    ec.addSectionProvider(
        registry.getProxyProvider(
            getListPropOrAll("ErrorCollector.section.providers"), HtmlReportPage.class));

    gp.addSectionProvider(
        registry.getProxyProvider(
            getListPropOrAll("FsmGraphProcessor.section.providers"), HtmlReportPage.class));

    setAutoRefs("error");
    setAutoRefs("fsm");
    setAutoRefs("coverage");
  }

  private String[] getListProp(String propName) {
    return config.getProperty(propName, "").split("\\s*,\\s*");
  }

  private String[] getListPropOrAll(String propName) {
    String propValue = config.getProperty(propName, "");
    if (propValue.equals("*")) {
      return ContextProviderRegistry.ALL;
    } else {
      return propValue.split("\\s*,\\s*");
    }
  }

  private void setAutoRefs(String targetAspect) {
    for (String asp : getListProp("autoref." + targetAspect)) {
      boolean biDirect;
      if (asp.endsWith("+")) {
        biDirect = true;
        asp = asp.substring(0, asp.length() - 1);
      } else {
        biDirect = false;
      }
      rg.addAutoRef(targetAspect, asp, biDirect);
    }
  }

  // ===============================================

  void generateFromTrace() {
    try {
      for (TraceProvider trace : traces) {
        tpm.addProvider(trace);
      }
      tpm.process();
    } catch (Exception ex) {
      ex.printStackTrace();
      System.err.println();
      System.err.println("Trying to generate report after failure...");
    }

    try {
      generateReport();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  void generateReport() throws IOException {
    System.err.println("Generating report...");

    rg.addIndexPage().setTitle("Report generated by MainApp");
    rg.addPages(tp.getPages());
    rg.addPages(ec.getPages());
    rg.addPages(cc.getPages());
    rg.addPages(oc.getPages());

    // А это не работает, потому что наша страница оракула не имеет чёткой
    // привязки к месту в трассе, а значит, не имеет контекста, за который
    // её можно было бы привязать к сценарию с помощью механизма ссылок
    // (однако, её можно привязать через механизм контекстной информации).
    // rg.addAutoRef("oracle", "fsm");

    // TODO: Сделать ссылки на вызов оракула только для покрытий аспекта impl
    // или для покрытий всех аспектов, кроме oracle
    // rg.addAutoRef("coverage", "oracle");

    rg.generateReport();

    System.err.println("Report generated successfully");
  }

  // ===============================================

  /**
   * Симуляция теста, генерирующего некую трассу. Метод вызывается при запуске с флагом "-sample".
   *
   * @param traceName Имя файла трассы или null.
   * @param onFly Надо ли генерировать на лету отчёт.
   */
  void generateTrace(String traceName, boolean onFly) {
    try {
      Tracer.setMasterTracer();

      if (traceName != null) {
        Tracer.getInstance()
            .addXmlTrace(new File(traceName), config.getProperty("trace.encoding", "utf-8"));
      }
      if (onFly) {
        /* Простой путь: трассировка напрямую в обработчики под управлением tpm */
        Tracer.getInstance().addProcessor(tpm);
      }

      TestTracer.traceTestStart("account_test.exe");
      final int fsmId = FsmTracer.traceFsmStart("account_scenario");

      CommonTracer.traceChannelDesc("0", "My test");
      Tracer.traceSyncTag("test", "start");

      FsmTracer.traceState(fsmId, 0, "0");
      FsmTracer.traceTransitionStart(fsmId, "deposit");
      FsmTracer.traceIterationVariable(fsmId, "int", "i", "1");
      final CoverageElement elA;
      final CoverageElement elB;
      final CoverageElement elU;
      final CoverageStructure depositCs =
          new CoverageStructure(
              "oracle",
              "marked path:deposit_spec(int)",
              new CoverageElement(
                  "branch Empty account",
                  new CoverageElement(
                      "Non-zero deposit",
                      elA = new CoverageElement("A"),
                      elB = new CoverageElement("B"),
                      elU = new CoverageElement("Unreachable")),
                  new CoverageElement("Zero deposit")),
              new CoverageElement("branch Non-empty account"));

      CoverageTracer.traceCoverageStructure(depositCs);
      OracleTracer.traceOracleStart("deposit_spec(int)", new Parameter("int", "i", "1"));
      OracleTracer.tracePreVerdict(true);
      CoverageTracer.traceCoverageElement(elA);
      CoverageTracer.traceCoverageElement(elB);
      CoverageTracer.traceElementImpossible(elU);
      MarkedPath mp = new MarkedPath(depositCs);
      mp.traceMark("Non-zero deposit");
      mp.traceMark("B");
      mp.traceBranch("Empty account");
      CoverageTracer.traceCoverageElement("impl", "deposit(int)", "5-10");
      CoverageTracer.traceCoverageElement("impl", "deposit(int)", "25-37");
      OracleTracer.tracePostVerdict(true);
      OracleTracer.traceOracleEnd("deposit_spec(int)");
      FsmTracer.traceTransitionEnd(fsmId, "deposit");

      FsmTracer.traceState(fsmId, 1, "1");
      FsmTracer.traceTransitionStart(fsmId, "withdraw");
      FsmTracer.traceIterationVariable(fsmId, "int", "i", "1");
      FsmTracer.traceTransitionEnd(fsmId, "withdraw");

      FsmTracer.traceState(fsmId, 0, "0");
      int transId = FsmTracer.traceTransitionStartEx(fsmId, "deposit");
      FsmTracer.traceIterationVariable(fsmId, "int", "i", "1");

      int callId =
          OracleTracer.traceOracleStartEx("deposit_spec(int)", new Parameter("int", "i", "1"));
      OracleTracer.tracePreVerdict(true);
      MarkedPath mp2 = new MarkedPath(depositCs);
      mp2.traceBranch("Non-empty account");
      CoverageTracer.traceCoverageElement("impl", "deposit(int)", "25-37");
      CoverageTracer.traceCoverageElement("impl", "deposit(int)", "45-99", "exception");
      CoverageTracer.traceCoverageElement("impl", "deposit(int)", "45-99");
      OracleTracer.traceFailure(OracleTracer.FAILURE_POST_VIOLATION, "Failed to deposit anything");
      OracleTracer.tracePostVerdict(false);
      OracleTracer.traceOracleEndEx("deposit_spec(int)", callId);
      FsmTracer.traceTransitionEndEx(fsmId, "deposit", transId);

      FsmTracer.traceState(fsmId, 0, "0");
      FsmTracer.traceFailure(
          FsmTracer.FAILURE_INDETERM,
          "Недетерминированное поведение сценарного КА."
              + "\nОжидаемое состояние: 1, полученное состояние: 0.");
      FsmTracer.traceSummary(
          fsmId, TraceUtils.createStringMap("transitions", "2", "non-determ", "1"));
      FsmTracer.traceBrowserMessage(fsmId, MessageLevel.ERROR, "Non-determinism...");
      FsmTracer.traceFsmEnd(fsmId);
      TestTracer.traceTestEnd("account_test.exe", false);

      if (onFly) {
        generateReport();
      }

    } catch (IOException ioe) {
      ioe.printStackTrace();
    } finally {
      Tracer.getInstance().endTrace();
    }
  }
}
