/*
 * Copyright 2009-2019 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.aspectrace.sampleApp;

import com.unitesk.aspectrace.AspectAdapter;
import com.unitesk.aspectrace.TagMessage;
import com.unitesk.aspectrace.html.HtmlReportPage;
import com.unitesk.aspectrace.html.ReportPage;
import com.unitesk.aspectrace.html.StaticResource;
import com.unitesk.aspectrace.ref.ContextProvider;
import com.unitesk.aspectrace.ref.ReferenceResolver;
import com.unitesk.aspectrace.test.TestListener;

import java.io.PrintWriter;
import java.util.LinkedHashSet;
import java.util.Properties;
import java.util.Set;

public class TestProcessor
    implements TestListener,
        ReferenceResolver<HtmlReportPage>,
        ContextProvider<HtmlReportPage>,
        ConfigurableGenerator {
  int testNum = 0;

  AspectAdapter adapter;
  String testName;

  public TestProcessor() {}

  public TestProcessor(Properties config) {
    configure(config);
  }

  // --------------- ConfigurableGenerator

  @Override
  public void configure(Properties config) {}

  @Override
  public ReportPage[] getPages() {
    if (allPages.isEmpty()) {
      return new ReportPage[0];
    }

    ReportPage[] rp = new ReportPage[allPages.size() + 1];
    allPages.toArray(rp);
    rp[rp.length - 1] = new StaticResource(MainApp.CSS_SRC, MainApp.CSS);
    return rp;
  }

  // --------------- TestListener

  @Override
  public void attachedToAdapter(String aspect, AspectAdapter adapter) {
    this.adapter = adapter;
  }

  @Override
  public void processTestStart(String testName) {
    SummaryCollector.setCurrentTestName(testName);
    curPage = new TestPage(adapter.getCurrentChannel(), testName, testNum);
    allPages.add(curPage);

    adapter.addProcessorTag("test." + testNum, null); // Hope there is only one test
    testNum++;
  }

  @Override
  public void processTestEnd(String testName, boolean verdict) {
    adapter.removeProcessorTag("test." + testNum, null);
    curPage.verdict = verdict;
    curPage = null;
  }

  @Override
  public void processTagChange(TagMessage msg) {}

  // --------------- HtmlReportPage

  protected TestPage curPage;
  protected Set<TestPage> allPages = new LinkedHashSet<TestPage>();

  protected static class TestPage extends HtmlReportPage {

    String testName;
    String tag;
    Boolean verdict;

    TestPage(String channel, String testName, int testNum) {
      this.testName = testName;
      tag = "test." + testNum;

      setAssociatedChannel(channel);
      setPath("test/" + testName + ".html"); // Let it be generated
      setTitle("Test " + testName);
      setStyle(MainApp.CSS);
    }

    @Override
    public void generatePageBody(PrintWriter pw) {
      pw.println("<h1>Test " + testName + "</h1>");
      if (verdict != null) {
        pw.println("Verdict: <b>" + verdict + "</b>");
      }
    }

    @Override
    public String generateRef(HtmlReportPage from, String mark) {
      if (tag.equals(mark)) {
        mark = null;
      }
      return super.generateRef(from, mark);
    }
  }

  // --------------- ContextProvider

  @Override
  public HtmlReportPage[] getRelated(String channel, final HtmlReportPage target) {
    if (curPage == null) {
      return null;
    }
    final TestPage testPage = curPage;
    return new HtmlReportPage[] {
      new HtmlReportPage() {
        @Override
        public void generatePageBody(PrintWriter pw) {
          pw.println("<table class=\"section\" width=\"100%\">");
          pw.println("<tr><th colspan=\"2\">Test suite");
          pw.println("<tr><td>Test<td>" + testPage.generateRef(target, null));
          pw.println("</table>");
        }
      }
    };
  }

  // --------------- ReferenceResolver

  // Resolve something to this
  @Override
  public HtmlReportPage resolve(String channel, String ref) {
    for (TestPage p : allPages) {
      String pgChan = p.getAssociatedChannel();
      if (ref.equals(p.tag)) {
        if (pgChan == null || channel == null || pgChan.equals(channel)) {
          return p;
        }
      }
    }
    return null;
  }
}
